"""Module to convert single agent plans to multi-agent plans with joint actions."""
import logging
import re
from pathlib import Path
from typing import List, Tuple

from pddl_plus_parser.models import Domain, ActionCall, Operator, JointActionCall, NOP_ACTION, Problem, State
from pddl_plus_parser.multi_agent.common import create_initial_state, apply_actions

PLAN_COMPONENT_REGEX = r"[\d+ : ]?\(([\w+\s?-]+)\)"


class PlanConverter:
    """Class that converts single agent plans to multi-agent plans with joint actions."""

    ma_domain: Domain
    logger: logging.Logger

    def __init__(self, ma_domain: Domain):
        self.ma_domain = ma_domain
        self.logger = logging.getLogger(__name__)

    def _extract_plan_actions(self, plan: str, agent_names: List[str]) -> List[Tuple[ActionCall, str]]:
        """Extracts the actions from the multi-agent plan.

        :param plan: the multi-agent plan.
        :param agent_names: the names of the agents.
        :return: the list tuples containing the action and the agent that executes it.
        """
        self.logger.info(f"Extracting the actions from the multi-agent plan:\n{plan}")
        matches = re.finditer(PLAN_COMPONENT_REGEX, plan, re.MULTILINE)
        plan_seq = []
        for match in matches:
            action_sequence = match.group(1)
            self.logger.debug(f"action sequence - {action_sequence}")
            action_components = action_sequence.lower().split()
            action_name = action_components[0]
            action_parameters = action_components[1:]
            # assuming that only one agent executes an action
            executing_agent = [param for param in action_parameters if param in agent_names][0]
            plan_seq.append((ActionCall(action_name, action_parameters), executing_agent))

        return plan_seq

    def _validate_well_defined_action_literals(self, combined_actions: List[ActionCall], next_action: Operator) -> bool:
        """Validates whether the actions' grounded literals are well-defined.

        We define a contradiction as the following:
            If a fluent and its negation exist in the same time, than it is considered a contradiction.

        Note: extending to numeric actions is easy when considering that two actions cannot change the same function at
        the same time.

        :param combined_actions: the currently constructed joint action.
        :param next_action: the new action to consider to add to the joint action.
        :return: whether the grounded literals are well-defined.
        """
        self.logger.debug("Validating that the literals are well-defined!")
        accumulated_add_effects = set()
        accumulated_delete_effects = set()

        for action_call in combined_actions:
            if action_call.name == NOP_ACTION:
                continue

            op = Operator(self.ma_domain.actions[action_call.name], self.ma_domain, action_call.parameters)
            op.ground()
            accumulated_add_effects.update([p.untyped_representation for p in op.grounded_add_effects])
            accumulated_delete_effects.update([p.untyped_representation for p in op.grounded_delete_effects])

        next_action_add_effects = [p.untyped_representation for p in next_action.grounded_add_effects]
        next_action_del_effects = [p.untyped_representation for p in next_action.grounded_delete_effects]

        return not (len(accumulated_add_effects.intersection(next_action_del_effects)) > 0 or
                    len(accumulated_delete_effects.intersection(next_action_add_effects)) > 0)

    def _validate_well_defined_joint_action(self, current_state: State,
                                            combined_actions: List[ActionCall], next_action: ActionCall,
                                            next_executing_agent: str, agent_names: List[str],
                                            should_validate_concurrency_constraint: bool = True) -> bool:
        """Validates if the joint action is well-defined.

        Note: the method in which the algorithm validates if a joint-action is well-defined is as follows:
            1. if the next action's executing agent already appears in the joint action,
                then the joint action is not well-defined.
            2. for each grounded literal in the joint action, if the preconditions contain contradictions,
                or the effects contain contradictions, then the joint action is not well-defined.

        We define a contradiction as the following:
            If a fluent and its negation exist in the same time, then it is considered a contradiction.
            Another example for a contradiction is a violation of the concurrency constraint which means, in our case,
            that more than one agent is interacting with an object.

        :param combined_actions: the joint action that is currently constructed.
        :param next_action: the new action to consider to add to the joint action.
        :param next_executing_agent: the agent that executes the new action.
        :param agent_names: the names of the agents.
        :param should_validate_concurrency_constraint: whether to validate the concurrency constraint.
        :return: whether the joint action with the new action is well-defined.
        """
        self.logger.info(f"Validating the joint action with the new action {str(next_action)}")
        next_agent_action_index = agent_names.index(next_executing_agent)
        if combined_actions[next_agent_action_index].name != NOP_ACTION:
            self.logger.debug("The agent already executes an action in the joint action")
            return False

        joint_action = JointActionCall(actions=combined_actions)
        if len(set(joint_action.joint_parameters).intersection(set(next_action.parameters))) > 0 \
                and should_validate_concurrency_constraint:
            self.logger.debug("The new action violates the concurrency constraint!")
            return False

        next_action_op = Operator(self.ma_domain.actions[next_action.name], self.ma_domain, next_action.parameters)
        next_action_op.ground()
        if not next_action_op.is_applicable(current_state):
            return False

        return self._validate_well_defined_action_literals(combined_actions, next_action_op)

    def _create_joint_actions(self, problem: Problem, plan_actions: List[Tuple[ActionCall, str]],
                              agent_names: List[str],
                              should_validate_concurrency_constraint: bool = True) -> List[JointActionCall]:
        """Creates the joint actions from the single agent actions.

        :param plan_actions: the single agent actions.
        :param agent_names: the names of the agents.
        :param should_validate_concurrency_constraint: whether to validate the concurrency constraint.
        :return: the joint actions.
        """
        self.logger.info("Creating the joint actions from the single agent action plan!")
        joint_actions = []
        current_state = create_initial_state(problem)
        while len(plan_actions) > 0:
            self.logger.debug("Initializing joint action to have only NOP for all the agents")
            joint_action = [ActionCall(NOP_ACTION, []) for _ in agent_names]
            action, agent = plan_actions.pop(0)
            joint_action[agent_names.index(agent)] = action

            if len(plan_actions) == 0:
                joint_actions.append(JointActionCall(joint_action))
                break

            next_action, next_executing_agent = plan_actions[0]
            while self._validate_well_defined_joint_action(current_state, joint_action, next_action,
                                                           next_executing_agent,
                                                           agent_names, should_validate_concurrency_constraint):
                joint_action[agent_names.index(next_executing_agent)] = plan_actions.pop(0)[0]

            self.logger.debug(f"Created the joint action {[str(action) for action in joint_action]}")
            current_state = apply_actions(self.ma_domain, current_state, [action for action in joint_action
                                                                          if action.name != NOP_ACTION])
            joint_actions.append(JointActionCall(joint_action))

        return joint_actions

    def convert_plan(self, problem: Problem, plan_file_path: Path, agent_names: List[str],
                     should_validate_concurrency_constraint: bool = True) -> List[JointActionCall]:
        """Converts a single agent plan to a multi-agent plan containing joint actions.

        :param problem: the problem that the plan solves.
        :param plan_file_path: the path to the original single agent plan.
        :param agent_names: the names of the agents that appear in the plan.
        :param should_validate_concurrency_constraint: whether to validate the concurrency constraint.
        :return: the list of joint actions describing the multi-agent plan.
        """
        with open(plan_file_path, "rt") as plan_file:
            plan_data = self._extract_plan_actions(plan_file.read(), agent_names)
            plan_actions = self._create_joint_actions(problem, plan_data, agent_names,
                                                      should_validate_concurrency_constraint)

        return plan_actions

    @staticmethod
    def export_plan(plan_file_path: Path, plan_actions: List[JointActionCall]):
        """Exports the multi-agent plan to a file.

        :param plan_file_path: the path to the file to export the plan to.
        :param plan_actions: the list of joint actions to export.
        """
        with open(plan_file_path, "wt") as plan_file:
            plan_file.writelines([f"{str(joint_action)}\n" for joint_action in plan_actions[:-1]])
            plan_file.write(str(plan_actions[-1]))



"""Module test for the single agent plan converter."""
import pytest
from pytest import fixture

from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser
from pddl_plus_parser.models import Domain, Problem
from pddl_plus_parser.multi_agent import PlanConverter
from tests.multi_agent_tests.consts import SOKOBAN_DOMAIN_FILE_PATH, SOKOBAN_UNPARSED_PLAN_PATH, \
    WOODWORKING_UNPARSED_PLAN_PATH, WOODWORKING_AGENT_NAMES, COMBINED_PROBLEM_PATH, \
    SOKOBAN_PROBLEM_PATH, COMBINED_DOMAIN_PATH, SOKOBAN_PROBLEM_WITH_INTERACTING_ACTIONS_PATH, \
    SOKOBAN_UNPARSED_PLAN_WITH_INTERACTING_ACTIONS_PATH

SOKOBAN_AGENT_NAMES = ["player-01", "player-02"]


@fixture()
def sokoban_domain() -> Domain:
    domain_parser = DomainParser(SOKOBAN_DOMAIN_FILE_PATH, partial_parsing=False)
    return domain_parser.parse_domain()


@fixture()
def sokoban_plan_converter(sokoban_domain: Domain) -> PlanConverter:
    return PlanConverter(sokoban_domain)


@fixture()
def woodworking_domain() -> Domain:
    domain_parser = DomainParser(COMBINED_DOMAIN_PATH, partial_parsing=False)
    return domain_parser.parse_domain()


@fixture()
def woodworking_problem(woodworking_domain: Domain) -> Problem:
    return ProblemParser(problem_path=COMBINED_PROBLEM_PATH, domain=woodworking_domain).parse_problem()


@fixture()
def sokoban_problem(sokoban_domain: Domain) -> Problem:
    return ProblemParser(problem_path=SOKOBAN_PROBLEM_PATH, domain=sokoban_domain).parse_problem()


@fixture()
def sokoban_problem_with_interacting_actions(sokoban_domain: Domain) -> Problem:
    return ProblemParser(problem_path=SOKOBAN_PROBLEM_WITH_INTERACTING_ACTIONS_PATH,
                         domain=sokoban_domain).parse_problem()


@fixture()
def woodworking_plan_converter(woodworking_domain: Domain) -> PlanConverter:
    return PlanConverter(woodworking_domain)


def test_convert_plan_does_not_remove_actions_from_original_plan(sokoban_plan_converter: PlanConverter,
                                                                 sokoban_problem: Problem):
    joint_actions = sokoban_plan_converter.convert_plan(sokoban_problem, SOKOBAN_UNPARSED_PLAN_PATH,
                                                        agent_names=SOKOBAN_AGENT_NAMES)
    assert sum([action.action_count for action in joint_actions]) == 28


def test_convert_plan_reduce_plan_length(sokoban_plan_converter: PlanConverter, sokoban_problem: Problem):
    joint_actions = sokoban_plan_converter.convert_plan(sokoban_problem, SOKOBAN_UNPARSED_PLAN_PATH,
                                                        agent_names=SOKOBAN_AGENT_NAMES)
    assert len(joint_actions) < 28


def test_convert_plan_does_not_have_joint_actions_with_same_agent(sokoban_plan_converter: PlanConverter,
                                                                  sokoban_problem: Problem):
    joint_actions = sokoban_plan_converter.convert_plan(
        sokoban_problem, SOKOBAN_UNPARSED_PLAN_PATH, agent_names=SOKOBAN_AGENT_NAMES)
    for joint_action in joint_actions:
        assert len([param for param in joint_action.joint_parameters if param == SOKOBAN_AGENT_NAMES[0]]) <= 1
        assert len([param for param in joint_action.joint_parameters if param == SOKOBAN_AGENT_NAMES[1]]) <= 1


def test_convert_plan_allows_fully_interacting_actions_to_be_executed_concurrently(
        sokoban_plan_converter: PlanConverter, sokoban_problem_with_interacting_actions: Problem):
    try:
        joint_actions = sokoban_plan_converter.convert_plan(
            sokoban_problem_with_interacting_actions, SOKOBAN_UNPARSED_PLAN_WITH_INTERACTING_ACTIONS_PATH,
            agent_names=SOKOBAN_AGENT_NAMES, should_validate_concurrency_constraint=False)
        for joint_action in joint_actions:
            print([str(action) for action in joint_action.actions])
    except Exception:
        pytest.fail("Exception raised when converting plan with fully interacting actions")


def test_convert_plan_does_not_remove_actions_from_original_plan_when_domain_contains_constants(
        woodworking_plan_converter: PlanConverter, woodworking_problem: Problem):
    joint_actions = woodworking_plan_converter.convert_plan(woodworking_problem, WOODWORKING_UNPARSED_PLAN_PATH,
                                                            agent_names=WOODWORKING_AGENT_NAMES)
    assert sum([action.action_count for action in joint_actions]) == 6


def test_convert_plan_reduce_plan_length_when_domain_contains_constants(
        woodworking_plan_converter: PlanConverter, woodworking_problem: Problem):
    joint_actions = woodworking_plan_converter.convert_plan(woodworking_problem, WOODWORKING_UNPARSED_PLAN_PATH,
                                                            agent_names=WOODWORKING_AGENT_NAMES)
    assert len(joint_actions) < 6


def test_convert_plan_does_not_have_joint_actions_with_same_agent_when_domain_contains_constants(
        woodworking_plan_converter: PlanConverter, woodworking_problem: Problem):
    joint_actions = woodworking_plan_converter.convert_plan(woodworking_problem, WOODWORKING_UNPARSED_PLAN_PATH,
                                                            agent_names=WOODWORKING_AGENT_NAMES)
    for joint_action in joint_actions:
        for agent_name in WOODWORKING_AGENT_NAMES:
            assert len([param for param in joint_action.joint_parameters if param == agent_name]) <= 1


def test_convert_plan_produces_readable_plan(woodworking_plan_converter: PlanConverter, woodworking_problem: Problem):
    joint_actions = woodworking_plan_converter.convert_plan(woodworking_problem, WOODWORKING_UNPARSED_PLAN_PATH,
                                                            agent_names=WOODWORKING_AGENT_NAMES)

    print()
    for joint_action in joint_actions:
        print(joint_action)


def test_convert_plan_when_removing_concurrency_constraint_creates_shorter_plan_than_if_we_do_not(
        woodworking_plan_converter: PlanConverter, woodworking_problem: Problem):
    short_joint_actions = woodworking_plan_converter.convert_plan(woodworking_problem,
                                                                  WOODWORKING_UNPARSED_PLAN_PATH,
                                                                  agent_names=WOODWORKING_AGENT_NAMES,
                                                                  should_validate_concurrency_constraint=False)
    longer_joint_actions = woodworking_plan_converter.convert_plan(woodworking_problem,
                                                                   WOODWORKING_UNPARSED_PLAN_PATH,
                                                                   agent_names=WOODWORKING_AGENT_NAMES)
    assert len(short_joint_actions) <= len(longer_joint_actions)

    print()
    for joint_action in short_joint_actions:
        print(joint_action)



from .action_call import ActionCall, JointActionCall, NOP_ACTION
from .conditional_effect import ConditionalEffect
from .numerical_expression import construct_expression_tree, calculate, evaluate_expression, NumericalExpressionTree
from .observation import ActionCall, Observation, ObservedComponent, MultiAgentObservation, MultiAgentComponent
from .pddl_action import Action
from .pddl_domain import Domain
from .pddl_function import PDDLFunction
from .pddl_object import PDDLObject, PDDLConstant
from .pddl_operator import Operator, NOPOperator
from .pddl_predicate import SignatureType, Predicate, GroundedPredicate
from .pddl_problem import Problem
from .pddl_state import State
from .pddl_type import PDDLType
from .universal_quantifier import UniversalQuantifiedPrecondition, UniversalQuantifiedEffect



"""Module to represent an observed trajectory. Compared to a trajectory, this contains only the observed data."""
from typing import List, Dict

from pddl_plus_parser.models.action_call import ActionCall, JointActionCall
from pddl_plus_parser.models.pddl_object import PDDLObject
from pddl_plus_parser.models.pddl_state import State


class ObservedComponent:
    """Class representing a single observed component."""

    previous_state: State
    grounded_action_call: ActionCall
    next_state: State

    def __init__(self, previous_state: State, call: ActionCall, next_state: State):
        self.previous_state = previous_state
        self.grounded_action_call = call
        self.next_state = next_state

    def __str__(self):
        return f"previous state: {self.previous_state.serialize()}\n" \
               f"operator: {str(self.grounded_action_call)}\n" \
               f"next state: {self.next_state.serialize()}"


class MultiAgentComponent:
    """class representing a multi-agent observed component."""

    previous_state: State
    grounded_joint_action: JointActionCall
    next_state: State

    def __init__(self, previous_state: State, joint_action: List[ActionCall], next_state: State):
        self.previous_state = previous_state
        self.grounded_joint_action = JointActionCall(joint_action)
        self.next_state = next_state

    def __str__(self):
        return f"previous state: {self.previous_state.serialize()}\n" \
               f"operators: {str(self.grounded_joint_action)}\n" \
               f"next state: {self.next_state.serialize()}"


class Observation:
    """Class representing an observed trajectory data."""

    components: List[ObservedComponent]
    grounded_objects: Dict[str, PDDLObject]

    def __init__(self):
        self.components = []
        self.grounded_objects = {}

    def add_problem_objects(self, objects: Dict[str, PDDLObject]) -> None:
        """Add the objects from the problem to the observation data.

        :param objects: the objects from the problem.
        """
        self.grounded_objects = objects

    def add_component(self, previous_state: State, call: ActionCall, next_state: State) -> None:
        """Add a new component to the observation data.

        :param previous_state: the state observed before the action call.
        :param call: the grounded action call.
        :param next_state: the state after the action was executed.
        """
        self.components.append(ObservedComponent(previous_state, call, next_state))


class MultiAgentObservation:
    """Class representing an observed multi-agent trajectory data."""

    components: List[MultiAgentComponent]
    grounded_objects: Dict[str, PDDLObject]
    agents_in_observation: List[str]

    def __init__(self, executing_agents: List[str]):
        self.components = []
        self.grounded_objects = {}
        self.agents_in_observation = executing_agents

    def add_problem_objects(self, objects: Dict[str, PDDLObject]) -> None:
        """Add the objects from the problem to the observation data.

        :param objects: the objects from the problem.
        """
        self.grounded_objects = objects

    def add_component(self, previous_state: State, joint_action: List[ActionCall], next_state: State) -> None:
        """Add a new component to the observation data.

        :param previous_state: the state observed before the action call.
        :param joint_action: the grounded joint action.
        :param next_state: the state after the action was executed.
        """
        self.components.append(MultiAgentComponent(previous_state, joint_action, next_state))



"""Model representing a universal quantifier in a PDDL+ action."""
from typing import Set

from pddl_plus_parser.models.pddl_predicate import Predicate
from pddl_plus_parser.models.pddl_type import PDDLType
from pddl_plus_parser.models.conditional_effect import ConditionalEffect


class UniversalQuantifiedPrecondition:
    """Class representing a universally quantified precondition."""

    quantified_parameter: str
    quantified_type: PDDLType
    positive_conditions: Set[Predicate]
    negative_conditions: Set[Predicate]

    def __init__(self, quantified_parameter: str, quantified_type: PDDLType, positive_conditions: Set[Predicate],
                 negative_conditions: Set[Predicate]):
        self.quantified_parameter = quantified_parameter
        self.quantified_type = quantified_type
        self.positive_conditions = positive_conditions
        self.negative_conditions = negative_conditions

    def __str__(self):
        positive_conditionals = "\n\t".join([cond.untyped_representation for cond in self.positive_conditions])
        negative_conditionals = "\n\t".join([f"(not {negative_cond.untyped_representation})" for
                                             negative_cond in self.negative_conditions])
        return f"(forall ({self.quantified_parameter} - {self.quantified_type.name})" \
               f"\n\t{positive_conditionals}\n\t{negative_conditionals})"


class UniversalQuantifiedEffect:
    """Class representing a universal quantifier in a PDDL+ action."""

    quantified_parameter: str
    quantified_type: PDDLType
    conditional_effect: ConditionalEffect

    def __init__(self, quantified_parameter: str, quantified_type: PDDLType, conditional_effect: ConditionalEffect):
        self.quantified_parameter = quantified_parameter
        self.quantified_type = quantified_type
        self.conditional_effect = conditional_effect

    def __str__(self):
        return f"(forall ({self.quantified_parameter} - {self.quantified_type.name})" \
               f"\n\t\t{str(self.conditional_effect)})"



from .multi_agent_domain_converter import MultiAgentDomainsConverter
from .multi_agent_problem_converter import MultiAgentProblemsConverter
from .multi_agent_trajectory_exporter import MultiAgentTrajectoryTriplet, MultiAgentTrajectoryExporter
from .single_agent_plan_converter import PlanConverter



#!/usr/bin/python3
import argparse
import random
from pathlib import Path
from typing import NoReturn

from pddl_plus_parser.problem_generators import get_problem_template

TEMPLATE_FILE_PATH = Path("counters_template.pddl")


def generate_instance(instance_name: str, num_counters: int, max_int: int) -> str:
    """Generate a single planning problem instance.

    :param instance_name: the name of the problem instance.
    :param num_counters: the number of counters in the problem.
    :param max_int: the maximal integer value.
    :return: the string representing the planning problem.
    """
    template = get_problem_template(TEMPLATE_FILE_PATH)
    template_mapping = {
        "instance_name": instance_name,
        "domain_name": "fo-counters-rnd",
        "counters_list": " ".join([f"c{i}" for i in range(num_counters)]),
        "counters_initial_values": "\n\t".join(
            [f"(= (value c{i}) {random.randint(0, max_int)})" for i in range(num_counters)]),
        "counters_rate_values": "\n\t".join([f"(= (rate_value c{i}) 0)" for i in range(num_counters)]),
        "max_int_value": max_int,
        "counters_final_values": "\n\t".join(
            [f"(<= (+ (value c{i}) 1) (value c{i + 1}))" for i in range(num_counters - 1)])
    }
    return template.substitute(template_mapping)


def parse_arguments() -> argparse.Namespace:
    """Parse the command line arguments."""
    parser = argparse.ArgumentParser(description="Generate counters planning instance")
    parser.add_argument("--min_counters", required=True, help="Minimal number of counters in the problems")
    parser.add_argument("--max_counters", required=True, help="Maximal number of counters in the problems")
    parser.add_argument("--max_int", required=True, help="Maximal integer value")
    parser.add_argument("--output_path", required=True,
                        help="The path to the output folder where the problems will be saved")
    args = parser.parse_args()
    return args


def generate_multiple_problems(
        min_counters: int, max_counters: int, max_int: int, output_folder: Path) -> NoReturn:
    """Generate multiple problems based on the input arguments.

    :param min_counters: the minimal number of counters possible in the problems.
    :param max_counters: the maximal number of counters possible in the problems.
    :param max_int: the maximal integer value.
    :param output_folder: the path to the output folder where the problems will be saved.
    """
    counters_range = [i for i in range(min_counters, max_counters + 1)]
    for num_counters in counters_range:
        print(f"Generating problem with {num_counters} counters")
        with open(output_folder / f"pfile{num_counters}_{random.randint(0, 100)}.pddl", "wt") as problem_file:
            problem_file.write(
                generate_instance(f"instance_{num_counters}_{max_int}", num_counters, max_int))


def main():
    args = parse_arguments()
    generate_multiple_problems(min_counters=int(args.min_counters),
                               max_counters=int(args.max_counters),
                               max_int=int(args.max_int),
                               output_folder=Path(args.output_path))


if __name__ == '__main__':
    main()



"""Module test for the grounded operator class."""
from typing import List, Dict, Set

from pytest import fixture, fail

from pddl_plus_parser.lisp_parsers import DomainParser, PDDLTokenizer, ProblemParser
from pddl_plus_parser.models import Domain, Action, Operator, GroundedPredicate, PDDLFunction, State, Problem
from tests.lisp_parsers_tests.consts import SPIDER_PROBLEM_PATH
from tests.models_tests.consts import TEST_HARD_NUMERIC_DOMAIN, TEST_NUMERIC_DOMAIN, SPIDER_DOMAIN_PATH, \
    NURIKABE_DOMAIN_PATH, NURIKABE_PROBLEM_PATH

TEST_LIFTED_SIGNATURE_ITEMS = ["?s", "?d", "?i", "?m"]
TEST_GROUNDED_ACTION_CALL = ["s1", "test_direction", "test_instrument", "test_mode"]

AGRICOLA_LIFTED_SIGNATURE_ITEMS = ["?w1", "?w2", "?wmax", "?r", "?i1", "?i2"]
AGRICOLA_GROUNDED_ACTION_CALL = ["noworker", "w1", "w2", "round1", "n1", "n2"]

SPIDER_START_DEALING_CALL = []
SPIDER_DEAL_CARD_CALL = ["card-d0-s1-v3", "card-d0-s0-v3", "deal-0", "card-d0-s3-v1", "pile-0"]
NURIKABE_START_PAINTING_CALL = ["pos-0-0", "g0", "n1", "n0"]
NURIKABE_MOVE_PAINTING_CALL = ["pos-2-0", "pos-3-0", "g1", "n1", "n0"]


@fixture()
def domain() -> Domain:
    domain_parser = DomainParser(TEST_HARD_NUMERIC_DOMAIN)
    return domain_parser.parse_domain()


@fixture()
def agricola_domain() -> Domain:
    domain_parser = DomainParser(TEST_NUMERIC_DOMAIN)
    return domain_parser.parse_domain()


@fixture()
def spider_domain() -> Domain:
    domain_parser = DomainParser(SPIDER_DOMAIN_PATH)
    return domain_parser.parse_domain()


@fixture()
def nurikabe_domain() -> Domain:
    domain_parser = DomainParser(NURIKABE_DOMAIN_PATH)
    return domain_parser.parse_domain()


@fixture()
def spider_problem(spider_domain: Domain) -> Problem:
    return ProblemParser(problem_path=SPIDER_PROBLEM_PATH, domain=spider_domain).parse_problem()


@fixture()
def nurikabe_problem(nurikabe_domain: Domain) -> Problem:
    return ProblemParser(problem_path=NURIKABE_PROBLEM_PATH, domain=nurikabe_domain).parse_problem()


@fixture()
def numeric_action(domain: Domain) -> Action:
    return domain.actions["take_image"]


@fixture()
def agricola_numeric_action(agricola_domain: Domain) -> Action:
    return agricola_domain.actions["take_food"]


@fixture()
def spider_unconditional_action(spider_domain: Domain) -> Action:
    return spider_domain.actions["start-dealing"]


@fixture()
def spider_conditional_action(spider_domain: Domain) -> Action:
    return spider_domain.actions["deal-card"]


@fixture()
def nurikabe_unconditional_action(nurikabe_domain: Domain) -> Action:
    return nurikabe_domain.actions["start-painting"]


@fixture()
def nurikabe_conditional_action(nurikabe_domain: Domain) -> Action:
    return nurikabe_domain.actions["move-painting"]


@fixture()
def operator(domain: Domain, numeric_action: Action) -> Operator:
    return Operator(numeric_action, domain, TEST_GROUNDED_ACTION_CALL)


@fixture()
def agricola_operator(agricola_domain: Domain, agricola_numeric_action: Action) -> Operator:
    return Operator(agricola_numeric_action, agricola_domain, AGRICOLA_GROUNDED_ACTION_CALL)


@fixture()
def spider_start_dealing_operator(spider_domain: Domain, spider_unconditional_action: Action) -> Operator:
    return Operator(spider_unconditional_action, spider_domain, SPIDER_START_DEALING_CALL)


@fixture()
def spider_deal_card_operator(spider_domain: Domain, spider_conditional_action: Action) -> Operator:
    return Operator(spider_conditional_action, spider_domain, SPIDER_DEAL_CARD_CALL)


@fixture()
def nurikabe_start_painting_operator(nurikabe_domain: Domain, nurikabe_unconditional_action: Action) -> Operator:
    return Operator(nurikabe_unconditional_action, nurikabe_domain, NURIKABE_START_PAINTING_CALL)


@fixture()
def nurikabe_move_painting_operator(nurikabe_domain: Domain, nurikabe_conditional_action: Action,
                                    nurikabe_problem: Problem) -> Operator:
    return Operator(nurikabe_conditional_action, nurikabe_domain, NURIKABE_MOVE_PAINTING_CALL,
                    problem_objects=nurikabe_problem.objects)


@fixture()
def complete_state_predicates(domain: Domain) -> Dict[str, Set[GroundedPredicate]]:
    calibrated_predicate = domain.predicates["calibrated"]
    on_board_predicate = domain.predicates["on_board"]
    supports_predicate = domain.predicates["supports"]
    power_on_predicate = domain.predicates["power_on"]
    pointing_predicate = domain.predicates["pointing"]
    return {
        "(calibrated ?i)": {
            GroundedPredicate(name="calibrated", signature=calibrated_predicate.signature,
                              object_mapping={"?i": "test_instrument"})},
        "(on_board ?i ?s)": [
            GroundedPredicate(name="on_board", signature=on_board_predicate.signature,
                              object_mapping={"?i": "test_instrument", "?s": "s1"})],
        "(supports ?i ?m)": {
            GroundedPredicate(name="supports", signature=supports_predicate.signature,
                              object_mapping={"?i": "test_instrument", "?m": "test_mode"})},
        "(power_on ?i)": {
            GroundedPredicate(name="power_on", signature=power_on_predicate.signature,
                              object_mapping={"?i": "test_instrument"})},
        "(pointing ?s ?d)": {
            GroundedPredicate(name="pointing", signature=pointing_predicate.signature,
                              object_mapping={"?s": "s1", "?d": "test_direction"})},
    }


@fixture()
def numeric_state_variables(domain: Domain) -> Dict[str, PDDLFunction]:
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    data_function = PDDLFunction(name="data", signature={
        "test_direction": domain.types["direction"],
        "test_mode": domain.types["mode"]
    })
    return {
        data_capacity_function.untyped_representation: data_capacity_function,
        data_function.untyped_representation: data_function,
    }


@fixture()
def previous_state_with_missing_numeric_fluent(domain: Domain,
                                               complete_state_predicates: Dict[str, Set[GroundedPredicate]],
                                               numeric_state_variables: Dict[str, PDDLFunction]):
    data_function = PDDLFunction(name="data", signature={
        "test_direction": domain.types["direction"],
        "test_mode": domain.types["mode"]
    })
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    numeric_state_variables[data_capacity_function.untyped_representation].set_value(18.3)
    numeric_state_variables[data_function.untyped_representation].set_value(5.3)
    return State(predicates=complete_state_predicates, fluents=numeric_state_variables)


@fixture()
def valid_previous_state(domain: Domain,
                         complete_state_predicates: Dict[str, Set[GroundedPredicate]],
                         numeric_state_variables: Dict[str, PDDLFunction]):
    data_function = PDDLFunction(name="data", signature={
        "test_direction": domain.types["direction"],
        "test_mode": domain.types["mode"]
    })
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    data_stored_function = PDDLFunction(name="data-stored", signature={})
    data_stored_function.set_value(10)
    numeric_state_variables[data_capacity_function.untyped_representation].set_value(18.3)
    numeric_state_variables[data_function.untyped_representation].set_value(5.3)
    return State(predicates=complete_state_predicates, fluents={
        **numeric_state_variables,
        data_stored_function.untyped_representation: data_stored_function
    })


def test_ground_predicates_creates_grounded_version_of_lifted_predicates_with_object_names_in_the_parameters(
        operator: Operator, numeric_action: Action):
    test_lifted_predicates = numeric_action.positive_preconditions
    test_parameters_map = {
        lifted_param: grounded_object for lifted_param, grounded_object in zip(
            TEST_LIFTED_SIGNATURE_ITEMS, TEST_GROUNDED_ACTION_CALL)
    }
    grounded_predicates = operator._ground_predicates(lifted_predicates=test_lifted_predicates,
                                                      parameters_map=test_parameters_map)

    expected_grounded_preconditions = ['(on_board test_instrument s1)',
                                       '(power_on test_instrument)',
                                       '(pointing s1 test_direction)',
                                       '(calibrated test_instrument)',
                                       '(supports test_instrument test_mode)']
    assert len(grounded_predicates) == len(test_lifted_predicates)
    assert sorted([p.untyped_representation for p in grounded_predicates]) == sorted(expected_grounded_preconditions)


def test_typed_action_call_returns_correct_string(operator: Operator):
    assert operator.typed_action_call == \
           "(take_image s1 - satellite test_direction - direction test_instrument - instrument test_mode - mode)"


def test_ground_predicates_when_domain_contains_constants_grounds_action_correctly(
        agricola_operator: Operator, agricola_numeric_action: Action):
    test_lifted_predicates = agricola_numeric_action.positive_preconditions
    test_parameters_map = {
        lifted_param: grounded_object for lifted_param, grounded_object in zip(
            AGRICOLA_LIFTED_SIGNATURE_ITEMS, AGRICOLA_GROUNDED_ACTION_CALL)
    }
    grounded_predicates = agricola_operator._ground_predicates(lifted_predicates=test_lifted_predicates,
                                                               parameters_map=test_parameters_map)

    expected_grounded_preconditions = ['(available_action act_labor)',
                                       '(current_worker noworker)',
                                       '(next_worker noworker w1)',
                                       '(max_worker w2)',
                                       '(current_round round1)',
                                       '(num_food n1)',
                                       '(next_num n1 n2)']
    assert len(grounded_predicates) == len(test_lifted_predicates)
    assert sorted([p.untyped_representation for p in grounded_predicates]) == sorted(expected_grounded_preconditions)


def test_ground_numeric_function_when_domain_contains_constants_grounds_action_correctly(
        agricola_operator: Operator, agricola_numeric_action: Action):
    test_grounded_call = ["w1", "w2", "noworker", "round1", "n1", "n2"]
    test_lifted_function = agricola_numeric_action.numeric_effects.pop()
    test_parameters_map = {
        lifted_param: grounded_object for lifted_param, grounded_object in zip(
            AGRICOLA_LIFTED_SIGNATURE_ITEMS, test_grounded_call)
    }
    grounded_expression_tree = agricola_operator._ground_numeric_calculation_tree(
        lifted_numeric_exp_tree=test_lifted_function, parameters_map=test_parameters_map)
    print(grounded_expression_tree)


def test_ground_predicates_creates_grounded_version_of_lifted_predicates_with_correct_parameter_mapping(
        operator: Operator, numeric_action: Action):
    test_lifted_predicates = numeric_action.positive_preconditions
    test_parameters_map = {
        lifted_param: grounded_object for lifted_param, grounded_object in zip(
            TEST_LIFTED_SIGNATURE_ITEMS, TEST_GROUNDED_ACTION_CALL)
    }
    grounded_predicates = operator._ground_predicates(lifted_predicates=test_lifted_predicates,
                                                      parameters_map=test_parameters_map)

    for predicate in grounded_predicates:
        if predicate.name == "calibrated":
            assert predicate.object_mapping == {"?i": "test_instrument"}


def test_ground_numeric_calculation_tree_extracts_correct_grounded_tree_data_from_lifted_calc_tree(
        operator: Operator, numeric_action: Action):
    test_lifted_expression_tree = numeric_action.numeric_preconditions.pop()
    test_parameters_map = {
        lifted_param: grounded_object for lifted_param, grounded_object in zip(
            TEST_LIFTED_SIGNATURE_ITEMS, TEST_GROUNDED_ACTION_CALL)
    }

    expression_tree = operator._ground_numeric_calculation_tree(test_lifted_expression_tree, test_parameters_map)

    root = expression_tree.root
    assert root.id == ">="
    assert root.children[0].id == "(data_capacity s1 - satellite)"
    assert root.children[1].id == "(data test_direction - direction test_mode - mode)"


def test_ground_equality_objects_returns_correct_grounded_objects(operator: Operator):
    equality_precondition = {("?size_before", "?size_after")}
    parameter_map = {"?size_before": "size1", "?size_after": "size2"}
    grounded_objects = operator._ground_equality_objects(equality_precondition, parameter_map)
    assert grounded_objects == {("size1", "size2")}


def test_equality_holds_when_objects_not_equal_returns_false(operator: Operator):
    assert not operator._equality_holds({("size1", "size2")})


def test_equality_holds_when_not_all_objects_equal_returns_false(operator: Operator):
    assert not operator._equality_holds({("size1", "size1"), ("size3", "size1")})


def test_equality_holds_when_objects_equal_returns_true(operator: Operator):
    assert operator._equality_holds({("size1", "size1")})


def test_ground_numeric_expressions_extracts_all_numeric_expressions(operator: Operator, numeric_action: Action):
    test_lifted_expression_tree = numeric_action.numeric_effects
    test_parameters_map = {
        lifted_param: grounded_object for lifted_param, grounded_object in zip(
            TEST_LIFTED_SIGNATURE_ITEMS, TEST_GROUNDED_ACTION_CALL)
    }

    expression_tree_set = operator._ground_numeric_expressions(test_lifted_expression_tree, test_parameters_map)

    assert len(expression_tree_set) == 2


def test_ground_grounds_positive_preconditions_with_correct_objects(operator: Operator):
    operator.ground()
    positive_grounded_preconditions = operator.grounded_positive_preconditions

    expected_grounded_preconditions = ['(on_board test_instrument s1)',
                                       '(power_on test_instrument)',
                                       '(pointing s1 test_direction)',
                                       '(calibrated test_instrument)',
                                       '(supports test_instrument test_mode)']

    assert len(positive_grounded_preconditions) == len(expected_grounded_preconditions)
    assert sorted([p.untyped_representation for p in positive_grounded_preconditions]) == sorted(
        expected_grounded_preconditions)


def test_ground_grounds_negative_preconditions_with_correct_objects(operator: Operator):
    operator.ground()
    positive_grounded_preconditions = operator.grounded_negative_preconditions

    expected_grounded_preconditions = []
    assert len(positive_grounded_preconditions) == len(expected_grounded_preconditions)


def test_ground_grounds_numeric_preconditions_with_correct_objects(operator: Operator):
    operator.ground()
    positive_grounded_numeric_preconditions = operator.grounded_numeric_preconditions

    assert len(positive_grounded_numeric_preconditions) == 1
    numeric_expression = positive_grounded_numeric_preconditions.pop()

    root = numeric_expression.root
    assert root.id == ">="
    assert root.children[0].id == "(data_capacity s1 - satellite)"
    assert root.children[1].id == "(data test_direction - direction test_mode - mode)"


def test_ground_grounds_numeric_effects_with_correct_objects(operator: Operator):
    operator.ground()
    positive_grounded_numeric_effects = operator.grounded_numeric_effects

    assert len(positive_grounded_numeric_effects) == 2

    first_expression = positive_grounded_numeric_effects.pop()
    second_expression = positive_grounded_numeric_effects.pop()

    expression_map = {
        first_expression.root.id: first_expression.root,
        second_expression.root.id: second_expression.root
    }

    root = expression_map["decrease"]
    assert root.children[0].id == "(data_capacity s1 - satellite)"
    assert root.children[1].id == "(data test_direction - direction test_mode - mode)"

    root = expression_map["increase"]
    assert root.children[0].id == "(data-stored )"
    assert root.children[1].id == "(data test_direction - direction test_mode - mode)"


def test_ground_grounds_boolean_effects_with_correct_objects(operator: Operator):
    operator.ground()
    grounded_add_effects = operator.grounded_add_effects

    assert len(grounded_add_effects) == 1
    expected_grounded_add_effect = ['(have_image test_direction test_mode)']
    assert [p.untyped_representation for p in grounded_add_effects] == expected_grounded_add_effect
    assert len(operator.grounded_delete_effects) == 0


def test_is_applicable_return_false_when_one_predicate_missing_in_state_predicates(
        operator: Operator, domain: Domain, complete_state_predicates: Dict[str, List[GroundedPredicate]],
        numeric_state_variables: Dict[str, PDDLFunction]):
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    numeric_state_variables[data_capacity_function.untyped_representation].set_value(5.3)

    missing_predicate_state_variables = {**complete_state_predicates}
    missing_predicate_state_variables.pop("(calibrated ?i)")

    state_with_missing_predicate = State(predicates=missing_predicate_state_variables, fluents=numeric_state_variables)
    assert not operator.is_applicable(state_with_missing_predicate)


def test_is_applicable_return_true_when_all_predicates_are_present_in_the_state(
        operator: Operator, domain: Domain, complete_state_predicates: Dict[str, Set[GroundedPredicate]],
        numeric_state_variables: Dict[str, PDDLFunction]):
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    numeric_state_variables[data_capacity_function.untyped_representation].set_value(5.3)
    state_with_complete_predicates = State(predicates=complete_state_predicates, fluents=numeric_state_variables)
    assert operator.is_applicable(state_with_complete_predicates)


def test_is_applicable_return_false_when_numeric_fluents_dont_correspond_preconditions_requirements(
        operator: Operator, domain: Domain, complete_state_predicates: Dict[str, Set[GroundedPredicate]],
        numeric_state_variables: Dict[str, PDDLFunction]):
    data_function = PDDLFunction(name="data", signature={
        "test_direction": domain.types["direction"],
        "test_mode": domain.types["mode"]
    })
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    numeric_state_variables[data_capacity_function.untyped_representation].set_value(5.3)
    numeric_state_variables[data_function.untyped_representation].set_value(18.7)
    state_with_complete_predicates = State(predicates=complete_state_predicates, fluents=numeric_state_variables)
    assert not operator.is_applicable(state_with_complete_predicates)


def test_is_applicable_return_true_when_numeric_fluents_are_equal(
        operator: Operator, domain: Domain, complete_state_predicates: Dict[str, Set[GroundedPredicate]],
        numeric_state_variables: Dict[str, PDDLFunction]):
    data_function = PDDLFunction(name="data", signature={
        "test_direction": domain.types["direction"],
        "test_mode": domain.types["mode"]
    })
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    numeric_state_variables[data_capacity_function.untyped_representation].set_value(5.3)
    numeric_state_variables[data_function.untyped_representation].set_value(5.3)
    state_with_complete_predicates = State(predicates=complete_state_predicates, fluents=numeric_state_variables)
    assert operator.is_applicable(state_with_complete_predicates)


def test_is_applicable_return_true_when_given_correct_fluents_with_negative_values(
        operator: Operator, domain: Domain, complete_state_predicates: Dict[str, Set[GroundedPredicate]],
        numeric_state_variables: Dict[str, PDDLFunction]):
    data_function = PDDLFunction(name="data", signature={
        "test_direction": domain.types["direction"],
        "test_mode": domain.types["mode"]
    })
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    numeric_state_variables[data_capacity_function.untyped_representation].set_value(-5.3)
    numeric_state_variables[data_function.untyped_representation].set_value(-15.32)
    state_with_complete_predicates = State(predicates=complete_state_predicates, fluents=numeric_state_variables)
    assert operator.is_applicable(state_with_complete_predicates)


def test_update_state_functions_does_not_raise_error_when_all_functions_are_present_in_state(
        operator: Operator, valid_previous_state: State):
    try:
        operator.ground()
        operator.update_state_functions(valid_previous_state)
    except Exception:
        fail()


def test_update_state_functions_returns_state_variables_with_correct_new_values(
        domain: Domain, operator: Operator, valid_previous_state: State):
    operator.ground()
    new_state_numeric_fluents = operator.update_state_functions(valid_previous_state)

    data_function = PDDLFunction(name="data", signature={
        "test_direction": domain.types["direction"],
        "test_mode": domain.types["mode"]
    })
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    data_stored_function = PDDLFunction(name="data-stored", signature={})

    assert new_state_numeric_fluents[data_function.untyped_representation].value == 5.3
    assert new_state_numeric_fluents[data_capacity_function.untyped_representation].value == 18.3 - 5.3
    assert new_state_numeric_fluents[data_stored_function.untyped_representation].value == 10 + 5.3


def test_update_state_functions_does_not_add_redundant_numeric_state_variables(
        operator: Operator, valid_previous_state: State):
    operator.ground()
    new_state_numeric_fluents = operator.update_state_functions(valid_previous_state)

    assert len(new_state_numeric_fluents) == 3


def test_update_state_predicates_adds_new_predicate_to_the_new_state(
        operator: Operator, valid_previous_state: State):
    operator.ground()
    new_state_predicates = operator.update_state_predicates(valid_previous_state)

    assert len(new_state_predicates) == len(valid_previous_state.state_predicates) + 1


def test_update_state_predicates_adds_the_correct_predicate(
        operator: Operator, valid_previous_state: State):
    operator.ground()
    new_state_predicates = operator.update_state_predicates(valid_previous_state)

    assert "(have_image ?d ?m)" in new_state_predicates


def test_update_state_predicates_removed_predicate_when_predicate_in_delete_effects(
        domain: Domain, operator: Operator, valid_previous_state: State):
    pointing_predicate_str = "(pointing ?s ?d)"
    assert len(valid_previous_state.state_predicates[pointing_predicate_str]) == 1
    pointing_predicate = domain.predicates["pointing"]

    operator.ground()
    operator.grounded_delete_effects = {
        GroundedPredicate(name="pointing", signature=pointing_predicate.signature,
                          object_mapping={"?s": "s1", "?d": "test_direction"})}

    new_state_predicates = operator.update_state_predicates(valid_previous_state)

    assert len(new_state_predicates[pointing_predicate_str]) == 0


def test_apply_returns_new_state_with_correct_values(
        domain: Domain, operator: Operator, valid_previous_state: State):
    next_state = operator.apply(valid_previous_state)
    next_state_fluents = next_state.state_fluents
    next_state_predicates = next_state.state_predicates

    assert "(have_image ?d ?m)" in next_state_predicates
    assert len(next_state_fluents) == 3
    data_function = PDDLFunction(name="data", signature={
        "test_direction": domain.types["direction"],
        "test_mode": domain.types["mode"]
    })
    data_capacity_function = PDDLFunction(name="data_capacity", signature={"s1": domain.types["satellite"]})
    data_stored_function = PDDLFunction(name="data-stored", signature={})

    assert next_state_fluents[data_function.untyped_representation].value == 5.3
    assert next_state_fluents[data_capacity_function.untyped_representation].value == 18.3 - 5.3
    assert next_state_fluents[data_stored_function.untyped_representation].value == 10 + 5.3


def test_is_applicable_with_disjunctive_action_operator_does_not_fail(valid_previous_state: State):
    test_action_str = """  (take_image
   :parameters (?s - satellite ?d - direction ?i - instrument ?m - mode)
   :precondition (and (calibrated ?i)
                      (on_board ?i ?s)
                      (supports ?i ?m)
                      (power_on ?i)
                      (pointing ?s ?d)
                      (power_on ?i)
                      (or 
                      (and (>= (data_capacity ?s) (data ?d ?m)))
			          (and (>= (data_capacity ?s) 0))
               ))
   :effect (and (decrease (data_capacity ?s) (data ?d ?m)) (have_image ?d ?m)
		(increase (data-stored) (data ?d ?m)))
  )"""
    domain_parser = DomainParser(TEST_HARD_NUMERIC_DOMAIN)
    domain = domain_parser.parse_domain()
    action_tokens = PDDLTokenizer(pddl_str=test_action_str).parse()
    action = domain_parser.parse_action(action_tokens, domain.types, domain.functions, domain.predicates,
                                        domain.constants)
    test_operator = Operator(action, domain, TEST_GROUNDED_ACTION_CALL)
    try:
        test_operator.is_applicable(valid_previous_state)
    except Exception:
        fail()


def test_apply_with_conditional_effects_does_not_fail(
        spider_domain: Domain,
        spider_start_dealing_operator: Operator,
        spider_deal_card_operator: Operator,
        spider_problem: Problem):
    try:
        initial_state = State(spider_problem.initial_state_predicates, spider_problem.initial_state_fluents)
        second_state = spider_start_dealing_operator.apply(initial_state)
        third_state = spider_deal_card_operator.apply(second_state)
        assert third_state is not None

    except Exception:
        fail()


def test_apply_with_conditional_effects_outputs_conditional_effects_in_successive_state(
        spider_domain: Domain,
        spider_start_dealing_operator: Operator,
        spider_deal_card_operator: Operator,
        spider_problem: Problem):
    initial_state = State(spider_problem.initial_state_predicates, spider_problem.initial_state_fluents)
    second_state = spider_start_dealing_operator.apply(initial_state)
    third_state = spider_deal_card_operator.apply(second_state)
    state_predicates = set()
    for predicates in third_state.state_predicates.values():
        state_predicates.update(predicates)

    assert "(currently-updating-unmovable )" in [p.untyped_representation for p in state_predicates]
    assert "(make-unmovable card-d0-s3-v1)" in [p.untyped_representation for p in state_predicates]


def test_apply_with_action_with_universal_quantifier_applies_effects_on_all_objects_that_match_conditions(
        nurikabe_domain: Domain, nurikabe_move_painting_operator: Operator, nurikabe_problem: Problem):
    initial_state = State(nurikabe_problem.initial_state_predicates, nurikabe_problem.initial_state_fluents)
    serialized_initial_state = initial_state.serialize()
    assert "(blocked pos-2-0)" not in serialized_initial_state
    assert "(blocked pos-4-0)" not in serialized_initial_state
    new_state = nurikabe_move_painting_operator.apply(initial_state, allow_inapplicable_actions=True)
    serialized_state = new_state.serialize()
    assert "(available pos-2-0)" not in serialized_state
    assert "(available pos-4-0)" not in serialized_state
    assert "(part-of pos-2-0 g1)" not in serialized_state
    assert "(part-of pos-4-0 g1)" not in serialized_state
    assert "(blocked pos-2-0)" in serialized_state
    assert "(blocked pos-4-0)" in serialized_state



"""Module test for the trajectory parser."""
from pytest import fixture

from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser, TrajectoryParser
from pddl_plus_parser.models import Domain, Problem
from tests.lisp_parsers_tests.consts import TEST_NUMERIC_DEPOT_DOMAIN, TEST_NUMERIC_DEPOT_PROBLEM, \
    TEST_NUMERIC_DEPOT_TRAJECTORY, FARMLAND_NUMERIC_DOMAIN, FARMLAND_NUMERIC_PROBLEM, FARMLAND_NUMERIC_TRAJECTORY, \
    WOODWORKING_COMBINED_DOMAIN_PATH, WOODWORKING_COMBINED_PROBLEM_PATH, WOODWORKING_COMBINED_TRAJECTORY_PATH, \
    DEPOT_MA_PROBLEM_PATH, DEPOT_MA_DOMAIN_PATH, DEPOT_MA_TRAJECTORY_PATH
from tests.multi_agent_tests.consts import WOODWORKING_AGENT_NAMES


@fixture()
def domain() -> Domain:
    domain_parser = DomainParser(TEST_NUMERIC_DEPOT_DOMAIN, partial_parsing=True)
    return domain_parser.parse_domain()


@fixture()
def problem(domain: Domain) -> Problem:
    return ProblemParser(problem_path=TEST_NUMERIC_DEPOT_PROBLEM, domain=domain).parse_problem()


@fixture()
def trajectory_parser(domain: Domain, problem: Problem) -> TrajectoryParser:
    return TrajectoryParser(domain, problem)


@fixture()
def farmland_domain() -> Domain:
    domain_parser = DomainParser(FARMLAND_NUMERIC_DOMAIN, partial_parsing=True)
    return domain_parser.parse_domain()


@fixture()
def farmland_problem(farmland_domain: Domain) -> Problem:
    return ProblemParser(problem_path=FARMLAND_NUMERIC_PROBLEM, domain=farmland_domain).parse_problem()


@fixture()
def farmland_trajectory_parser(farmland_domain: Domain, farmland_problem: Problem) -> TrajectoryParser:
    return TrajectoryParser(farmland_domain, farmland_problem)


@fixture()
def ma_combined_domain() -> Domain:
    return DomainParser(WOODWORKING_COMBINED_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def ma_combined_problem(ma_combined_domain: Domain) -> Problem:
    return ProblemParser(problem_path=WOODWORKING_COMBINED_PROBLEM_PATH, domain=ma_combined_domain).parse_problem()


@fixture()
def ma_trajectory_parser(ma_combined_domain: Domain, ma_combined_problem: Problem) -> TrajectoryParser:
    return TrajectoryParser(ma_combined_domain, ma_combined_problem)


@fixture()
def ma_depot_domain() -> Domain:
    return DomainParser(DEPOT_MA_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def ma_depot_problem(ma_depot_domain: Domain) -> Problem:
    return ProblemParser(problem_path=DEPOT_MA_PROBLEM_PATH, domain=ma_depot_domain).parse_problem()


@fixture()
def ma_depot_trajectory_parser(ma_depot_domain: Domain, ma_depot_problem: Problem) -> TrajectoryParser:
    return TrajectoryParser(ma_depot_domain, ma_depot_problem)


def test_parse_trajectory(trajectory_parser: TrajectoryParser):
    observation = trajectory_parser.parse_trajectory(TEST_NUMERIC_DEPOT_TRAJECTORY)
    assert len(observation.components) == 19


def test_parse_farmland_trajectory(farmland_trajectory_parser: TrajectoryParser):
    observation = farmland_trajectory_parser.parse_trajectory(FARMLAND_NUMERIC_TRAJECTORY)
    assert len(observation.components) == 12


def test_parse_ma_combined_trajectory(ma_trajectory_parser: TrajectoryParser):
    observation = ma_trajectory_parser.parse_trajectory(WOODWORKING_COMBINED_TRAJECTORY_PATH,
                                                        executing_agents=WOODWORKING_AGENT_NAMES)
    assert len(observation.components) == 4
    print(observation.components[0].grounded_joint_action)


def test_parse_depot_combined_trajectory(ma_depot_trajectory_parser: TrajectoryParser):
    observation = ma_depot_trajectory_parser.parse_trajectory(DEPOT_MA_TRAJECTORY_PATH,
                                                              executing_agents=["depot0", "distributor0",
                                                                                "distributor1",
                                                                                "distributor2", "distributor3",
                                                                                "truck0",
                                                                                "truck1", "truck2", "truck3"])
    assert "(clear pallet6)" not in observation.components[0].next_state.serialize()



"""Module to handle the functionality of conditional effects."""
from typing import Union, Set

from .pddl_predicate import Predicate, GroundedPredicate
from .numerical_expression import NumericalExpressionTree


class ConditionalEffect:
    """Class representing a conditional effect in a PDDL+ action."""

    positive_conditions: Set[Union[Predicate, GroundedPredicate]]
    negative_conditions: Set[Union[Predicate, GroundedPredicate]]
    numeric_conditions: Set[NumericalExpressionTree]
    add_effects: Set[Union[Predicate, GroundedPredicate]]
    delete_effects: Set[Union[Predicate, GroundedPredicate]]
    numeric_effects: Set[NumericalExpressionTree]

    def __init__(self):
        self.positive_conditions = set()
        self.negative_conditions = set()
        self.numeric_conditions = set()
        self.add_effects = set()
        self.delete_effects = set()
        self.numeric_effects = set()

    def __str__(self):
        positive_antecedents = "\n\t".join([cond.untyped_representation for cond in self.positive_conditions])
        negative_antecedents = "\n\t".join(
            [f"(not {negative_cond.untyped_representation})" for negative_cond in self.negative_conditions])
        numeric_antecedents = "\n\t".join([cond.to_pddl() for cond in self.numeric_conditions])
        add_effect = "\n\t".join([effect.untyped_representation for effect in self.add_effects])
        delete_effect = "\n\t".join(
            [f"(not {negative_effect.untyped_representation})" for negative_effect in self.delete_effects])
        discrete_effect = add_effect + delete_effect
        numeric_effect = "\n\t".join([effect.to_pddl() for effect in self.numeric_effects])

        return f"(when (and {positive_antecedents}{negative_antecedents}{numeric_antecedents}) " \
               f"(and {discrete_effect}{numeric_effect}))"



"""Module to parse a serialized trajectory and export it as an object fitted for learning."""
import logging
from collections import defaultdict
from pathlib import Path
from typing import List, Union

from pddl_plus_parser.lisp_parsers.pddl_tokenizer import PDDLTokenizer
from pddl_plus_parser.models import Domain, Observation, State, ActionCall, PDDLFunction, Predicate, GroundedPredicate, \
    Problem, MultiAgentObservation, NOP_ACTION


class TrajectoryParser:
    """Class that is able to parse a trajectory file and extract its content.

    Note: This class assumes that the trajectory size is tractable and thus, reads the entire file at once.
    For lay loading of a trajectory future development is needed.
    """

    partial_domain: Domain  # domain containing only the publicly known information.
    problem: Problem
    logger: logging.Logger

    def __init__(self, partial_domain: Domain, problem: Problem):
        self.partial_domain = partial_domain
        self.problem = problem
        self.logger = logging.getLogger(__name__)

    def _read_trajectory_file(self, trajectory_file_path: Path) -> PDDLTokenizer:
        """Reads the trajectory file and exports the lines containing the data about the trajectory.

        :param trajectory_file_path: the path to the trajectory file.
        :return: the tokenizer that is able to parse the trajectory strings to expressions.
        """
        self.logger.debug(f"Reading the file - {trajectory_file_path}")
        return PDDLTokenizer(file_path=trajectory_file_path)

    def parse_state(self, state_data: List[List[Union[str, List[str]]]]) -> State:
        """Parse the trajectory's state data and extracts the state.

        :param state_data: the AST representing the state.
        :return: the state object.
        """
        self.logger.info("Parsing the observed state.")
        state_predicates = defaultdict(set)
        state_fluents = {}
        for expression in state_data:
            if expression[0] == "=":  # This is an assignment of a grounded numeric fluent.
                if len(expression) != 3:  # ['=', <function items as a list>, '<value>']
                    raise SyntaxError("A numeric fluent should be of length 3. Fluent scheme: "
                                      "(= (<fluent_name> <argument>) <value>)"
                                      f"Received - {expression}")

                self.logger.debug("Component found is a numeric fluent, starting to process the fluent.")
                function_data = expression[1]
                assigned_value = float(expression[2])
                numeric_fluent = self.parse_grounded_numeric_fluent(function_data)
                self.logger.debug(f"Setting the fluent's value to - {assigned_value}")
                numeric_fluent.set_value(assigned_value)
                state_fluents[numeric_fluent.untyped_representation] = numeric_fluent
                continue

            if expression[0] in self.partial_domain.predicates:
                self.logger.debug("Component found is a predicate, starting to process it.")
                lifted_predicate = self.partial_domain.predicates[expression[0]]
                grounded_predicate = self.parse_grounded_predicate(expression, lifted_predicate)
                state_predicates[lifted_predicate.untyped_representation].add(grounded_predicate)
                continue

            raise ValueError(f"Received illegal state component - {expression}")

        return State(predicates=state_predicates, fluents=state_fluents)

    def parse_grounded_numeric_fluent(self, grounded_numeric_fluent: List[str]) -> PDDLFunction:
        """Parse a single grounded numeric fluent in the problem.

        :param grounded_numeric_fluent: the grounded fluent that is present in the problem.
        :return: the function object representing the grounded fluent.
        """
        function_name = grounded_numeric_fluent[0]
        self.logger.info(f"Starting to parse the grounded numeric fluent - {function_name}")
        assert function_name in self.partial_domain.functions
        lifted_function = self.partial_domain.functions[function_name]
        # For now, assuming that fluents have valid parameters.
        fluent_signature_items = grounded_numeric_fluent[1:]
        if len(fluent_signature_items) != len(lifted_function.signature):
            raise ValueError(
                f"Received fluent - {function_name} with wrong number of parameters! "
                f"Expected - {len(lifted_function.signature)} and received - {len(fluent_signature_items)}")

        possible_objects = {**self.problem.objects, **self.partial_domain.constants}
        fluent_signature = {
            object_name: possible_objects[object_name].type for object_name in fluent_signature_items
        }
        for grounded_param_type, lifted_param_type in zip(fluent_signature.values(),
                                                          lifted_function.signature.values()):
            assert grounded_param_type.is_sub_type(lifted_param_type)

        return PDDLFunction(name=function_name, signature=fluent_signature)

    def parse_grounded_predicate(self, grounded_predicate_ast: List[str],
                                 lifted_predicate: Predicate) -> GroundedPredicate:
        """Parse the grounded predicate that appears in the problem definition.

        :param grounded_predicate_ast: the AST that represents the grounded predicate.
        :param lifted_predicate: the lifted predicate that the domain defines.
        :return: the grounded predicate represented by the lifted predicate.
        """
        predicate_name = lifted_predicate.name
        predicate_signature_items = grounded_predicate_ast[1:]
        self.logger.info(f"Starting the parse the grounded predicate - {predicate_name} "
                         f"with the signature - {predicate_signature_items}.")
        if len(predicate_signature_items) != len(lifted_predicate.signature):
            raise ValueError(
                f"Received illegal grounded predicate with mismatching signature - {grounded_predicate_ast}")

        object_mapping = {
            parameter_name: object_name for object_name, parameter_name in
            zip(predicate_signature_items, lifted_predicate.signature)
        }

        return GroundedPredicate(name=predicate_name, signature=lifted_predicate.signature,
                                 object_mapping=object_mapping)

    def parse_action_call(self, action_call_ast: List[List[str]]) -> ActionCall:
        """Parse the grounded action call in the trajectory.

        :param action_call_ast: a grounded action call in the form: [(<name> <p1> <p2> ... <pn>)]
        :return: the action call object.
        """
        self.logger.debug(f"Parsing the grounded action call - {action_call_ast}")
        action_call_data = action_call_ast[0]
        return ActionCall(name=action_call_data[0], grounded_parameters=action_call_data[1:])

    def parse_joint_action(self, joint_action_call_ast: List[List[str]],
                           executing_agents: List[str]) -> List[ActionCall]:
        """Parse the joint action call AST in the trajectory.

        :param joint_action_call_ast: the joint action call AST in the form.
        :return: the joint action call object.
        """
        self.logger.debug(f"Parsing the joint action call - {joint_action_call_ast}")
        actions = []
        for agent_name, action_call_ast in zip(executing_agents, joint_action_call_ast):
            if action_call_ast[0] == NOP_ACTION:
                self.logger.debug(f"{agent_name} is executing a NOP action.")
                actions.append(ActionCall(name=NOP_ACTION, grounded_parameters=[]))
                continue

            actions.append(self.parse_action_call([action_call_ast]))

        return actions

    def parse_trajectory(self, trajectory_file_path: Path,
                         executing_agents: List[str] = None) -> Union[Observation, MultiAgentObservation]:
        """Parse a trajectory and extracts the observed data into objects.

        :param trajectory_file_path: the path to the trajectory file.
        :param executing_agents: the list of agents that partake in the observation.
        :return: the observation extracted from the serialized trajectory.
        """
        self.logger.info("Starting to read the trajectory file!")
        tokenizer = self._read_trajectory_file(trajectory_file_path)
        observation_expression = tokenizer.parse()
        observation = MultiAgentObservation(
            executing_agents=executing_agents) if executing_agents is not None else Observation()

        observation.add_problem_objects(self.problem.objects)
        self.logger.debug("Starting to generate the observation from the input trajectory.")
        for index in range(0, len(observation_expression) - 2, 2):
            macro_expression = observation_expression[index]
            if macro_expression[0] == ":init" or macro_expression[0] == ":state":
                previous_state = self.parse_state(macro_expression[1:])
                macro_expression = observation_expression[index + 1]
                if macro_expression[0] == "operator:":
                    action_call = self.parse_action_call(macro_expression[1:])

                elif macro_expression[0] == "operators:":
                    action_call = self.parse_joint_action(macro_expression[1:], executing_agents)

                else:
                    raise SyntaxError("Encountered a trajectory without an action call!")

                macro_expression = observation_expression[index + 2]
                if macro_expression[0] != ":state":
                    raise SyntaxError("Encountered a trajectory without a next state!")

                next_state = self.parse_state(macro_expression[1:])
                self.logger.debug("Finished parsing a trajectory component.")
                observation.add_component(previous_state, action_call, next_state)

        return observation



"""Class that represents a numerical expression that can be evaluated."""
from typing import List, Union, Dict, Optional

from anytree import AnyNode, RenderTree

from .pddl_function import PDDLFunction

LEGAL_NUMERICAL_EXPRESSIONS = ["=", "!=", "<=", ">=", ">", "<", "+", "-", "/", "*", "increase", "decrease", "assign"]


def construct_expression_tree(expression_ast: List[Union[str, List[str]]],
                              domain_functions: Dict[str, PDDLFunction]) -> AnyNode:
    """Constructs a tree that represents the numerical expression and that is later on able to be evaluated.

    :param expression_ast: the AST representing the numeric expression that is to be parsed.
    :param domain_functions: the functions that are defined in the domain.
    :return: the tree representing the numeric calculation.
    """
    if type(expression_ast) == str:
        if expression_ast in LEGAL_NUMERICAL_EXPRESSIONS:
            raise SyntaxError("Leaf node cannot be a numerical operator!")

        ast_node_item: str = expression_ast
        try:
            return AnyNode(id=f"{float(ast_node_item)}", value=float(ast_node_item))

        except ValueError:
            raise SyntaxError("Leaf node with bad string was encountered!")

    # This means that we have a list as a leaf --> a function that we need to create.
    elif all([type(item) == str for item in expression_ast]):
        function_name = expression_ast[0]
        extracted_function = domain_functions[function_name]
        if len(expression_ast) == 1:
            return AnyNode(id=str(extracted_function), value=extracted_function)

        new_function = PDDLFunction(name=function_name, signature={
            param_name: param_type for param_name, param_type in
            zip(expression_ast[1:], extracted_function.signature.values())
        })
        return AnyNode(id=str(new_function), value=new_function)

    node = AnyNode(id=expression_ast[0], value=expression_ast[0], children=[
        construct_expression_tree(expression_ast[1], domain_functions),
        construct_expression_tree(expression_ast[2], domain_functions)
    ])
    return node


def increase(value_to_increase: PDDLFunction, increase_by: float) -> None:
    """Increase the value of the first numerical fluent by the value of the other.

    :param value_to_increase: the parameter that is to be increased
    :param increase_by: the value to increase the parameter by.
    """
    previous_value = value_to_increase.value
    value_to_increase.set_value(previous_value + increase_by)


def decrease(value_to_decrease: PDDLFunction, decrease_by: float) -> None:
    """Decrease the value of the first numerical fluent by the value of the other.

    :param value_to_decrease: the parameter that is to be decreased
    :param decrease_by: the value to decrease the parameter by.
    """
    previous_value = value_to_decrease.value
    value_to_decrease.set_value(previous_value - decrease_by)


def assign(assigned_variable: PDDLFunction, value_to_assign: float) -> None:
    """assigns the value of one numeric variable to another variable.

    :param assigned_variable: the variable that is being assigned a new value.
    :param value_to_assign: the value that is being assigned to the input variable.
    """
    assigned_variable.set_value(value_to_assign)


COMPARISON_OPERATORS = {
    "=": lambda x, y: x == y,
    "!=": lambda x, y: x != y,
    "<=": lambda x, y: x <= y,
    ">=": lambda x, y: x >= y,
    ">": lambda x, y: x > y,
    "<": lambda x, y: x < y,
}

ASSIGNMENT_EXPRESSIONS = {
    "increase": increase,
    "decrease": decrease,
    "assign": assign
}

NUMERICAL_BINARY_OPERATORS = {
    "+": lambda x, y: x + y,
    "-": lambda x, y: x - y,
    "/": lambda x, y: x / y,
    "*": lambda x, y: x * y,
}


def calculate(expression_node: AnyNode) -> float:
    """Calculates the numerical value of an expression tree.

    :param expression_node: the node that is currently being observed.
    :return: the value that was calculated.
    """
    if len(expression_node.children) == 0:
        if isinstance(expression_node.value, PDDLFunction):
            node_function: PDDLFunction = expression_node.value
            return node_function.value

        return expression_node.value

    numerical_operator = expression_node.value
    left_operand = calculate(expression_node.children[0])
    right_operand = calculate(expression_node.children[1])
    return NUMERICAL_BINARY_OPERATORS[numerical_operator](left_operand, right_operand)


def evaluate_expression(expression_tree: AnyNode) -> Optional[Union[bool, PDDLFunction]]:
    """Evaluates the value of a PDDL expression based on parsing the content of its AST.

    :param expression_tree: the PDDL expression to evaluate.
    :return: bool if the operand is a comparison operand, otherwise assign the calculated value to the function inline.
    """
    if expression_tree.value in ASSIGNMENT_EXPRESSIONS:
        assigned_variable: PDDLFunction = expression_tree.children[0].value
        evaluated_operand = calculate(expression_tree.children[1])
        ASSIGNMENT_EXPRESSIONS[expression_tree.value](assigned_variable, evaluated_operand)
        return assigned_variable

    compared_operator = calculate(expression_tree.children[0])
    evaluated_operand = calculate(expression_tree.children[1])
    return COMPARISON_OPERATORS[expression_tree.value](compared_operator, evaluated_operand)


class NumericalExpressionTree:
    root: AnyNode

    def __init__(self, expression_tree: AnyNode):
        self.root = expression_tree

    def __str__(self):
        return "\n".join([f"{pre}{node.id}" for pre, _, node in RenderTree(self.root)])

    def _convert_to_pddl(self, node: AnyNode) -> str:
        """Recursive method that converts the expression tree to a PDDL string.

        :param node: the node that the recursion is currently working on.
        :return: the PDDL string of the expression.
        """
        if node.is_leaf:
            if isinstance(node.value, PDDLFunction):
                function: PDDLFunction = node.value
                return function.untyped_representation

            return node.value

        left_operand = self._convert_to_pddl(node.children[0])
        right_operand = self._convert_to_pddl(node.children[1])
        return f"({node.value} {left_operand} {right_operand})"

    def to_pddl(self) -> str:
        return self._convert_to_pddl(self.root)



from .ff_output_parser import MetricFFParser
from .numeric_trajectory_exporter import TrajectoryExporter, TrajectoryTriplet
from .enhsp_output_parser import ENHSPParser
from .domain_exporter import DomainExporter
from .problem_exporter import ProblemExporter



"""module to test the problem parsing functionality."""
from pytest import fixture, raises, fail

from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser, PDDLTokenizer
from pddl_plus_parser.models import Domain
from tests.lisp_parsers_tests.consts import TEST_NUMERIC_DOMAIN, TEST_NUMERIC_PROBLEM, ZENOTRAVEL_DOMAIN_PATH, \
    ZENOTRAVEL_PROBLEM_PATH

test_objects_ast = ['num0', 'num1', 'num2', 'num3', 'num4', 'num5', 'num6', 'num7', 'num8', 'num9', 'num10', 'num11',
                    'num12', 'num13', 'num14', 'num15', 'num16', '-', 'num', 'stage1', 'stage2', 'stage3', 'stage4',
                    '-', 'stage', 'round1', 'round2', 'round3', 'round4', 'round5', 'round6', 'round7', 'round8',
                    'round9', 'round10', '-', 'round', 'worker1', 'worker2', 'worker3', 'worker4', '-', 'worker',
                    'room1', 'room2', 'room3', 'room4', '-', 'room']


@fixture(scope="module")
def domain() -> Domain:
    domain_parser = DomainParser(TEST_NUMERIC_DOMAIN)
    return domain_parser.parse_domain()


@fixture()
def zenotravel_domain() -> Domain:
    domain_parser = DomainParser(ZENOTRAVEL_DOMAIN_PATH)
    return domain_parser.parse_domain()


@fixture()
def zenotravel_problem_parser(zenotravel_domain: Domain):
    return ProblemParser(problem_path=ZENOTRAVEL_PROBLEM_PATH, domain=zenotravel_domain)


@fixture(scope="module")
def problem_parser(domain: Domain) -> ProblemParser:
    return ProblemParser(problem_path=TEST_NUMERIC_PROBLEM, domain=domain)


def set_objects_in_problem(problem_parser):
    parsed_objects = problem_parser.parse_objects(test_objects_ast)
    problem_parser.problem.objects = parsed_objects


@fixture()
def problem_parser_with_objects(domain: Domain) -> ProblemParser:
    problem_parser = ProblemParser(problem_path=TEST_NUMERIC_PROBLEM, domain=domain)
    parsed_objects = problem_parser.parse_objects(test_objects_ast)
    problem_parser.problem.objects = parsed_objects
    return problem_parser


def test_parse_domain_name_when_given_invalid_domain_name_raises_error(problem_parser):
    bad_domain_name = "bad-domain-name"
    with raises(ValueError):
        problem_parser.parse_domain_name(bad_domain_name)


def test_parse_domain_name_when_given_valid_domain_name_no_error_is_raised(
        problem_parser, domain: Domain):
    try:
        problem_parser.parse_domain_name(domain.name)
    except ValueError:
        fail()


def test_parse_objects_parse_simple_objects_list_with_one_type_correctly(problem_parser):
    test_objects_ast = ['num1', 'num2', 'num3', 'num4', 'num5', 'num6', 'num7', 'num8', 'num9', 'num10', 'num11',
                        'num12', 'num13', 'num14', 'num15', 'num16', '-', 'num']
    parsed_objects = problem_parser.parse_objects(test_objects_ast)
    assert len(parsed_objects) == 16
    assert all([obj.type.name == "num" for obj in parsed_objects.values()])


def test_parse_objects_with_private_objects_parse_correctly_correctly(problem_parser):
    objects_str_with_private_objs = """(
    num0 num1 num2 num3  - num
    (:private
        stage1 stage2 - stage
        round1 round2  - round
    )
    worker1 worker2 - worker
    )
    """
    tokenizer = PDDLTokenizer(pddl_str=objects_str_with_private_objs)
    parsed_objects = problem_parser.parse_objects(tokenizer.parse())
    assert set(parsed_objects.keys()) == {"num0", "num1", "num2", "num3", "stage1", "stage2", "round1", "round2",
                                          "worker1", "worker2"}


def test_parse_objects_parse_objects_list_with_multiple_types_correctly(problem_parser):
    parsed_objects = problem_parser.parse_objects(test_objects_ast)
    objects_types = set([obj.type.name for obj in parsed_objects.values()])
    assert len(objects_types) == 5


def test_parse_grounded_numeric_fluent_with_different_number_of_parameters_than_domain_definition_raises_error(
        problem_parser):
    test_bad_function_ast = ['group_worker_cost', 'worker2', 'worker3']
    with raises(ValueError):
        problem_parser.parse_grounded_numeric_fluent(test_bad_function_ast)


def test_parse_grounded_numeric_fluent_with_correct_number_parameters_and_bad_parameter_type_raises_error(
        problem_parser_with_objects: ProblemParser):
    test_bad_function_ast = ['group_worker_cost', 'num8']
    with raises(AssertionError):
        problem_parser_with_objects.parse_grounded_numeric_fluent(test_bad_function_ast)


def test_parse_grounded_numeric_fluent_wrong_function_name_raises_error(
        problem_parser_with_objects: ProblemParser):
    test_bad_function_ast = ['group_worker_cost-blah', 'num8']
    with raises(AssertionError):
        problem_parser_with_objects.parse_grounded_numeric_fluent(test_bad_function_ast)


def test_parse_grounded_numeric_fluent_with_valid_grounded_fluent_definition_return_the_function_object_generated(
        problem_parser_with_objects: ProblemParser, domain: Domain):
    valid_grounded_function = ['group_worker_cost', 'worker2']
    extracted_function = problem_parser_with_objects.parse_grounded_numeric_fluent(valid_grounded_function)
    assert extracted_function.name == "group_worker_cost"
    assert extracted_function.signature == {
        "worker2": domain.types["worker"]
    }


def test_parse_grounded_numeric_fluent_with_valid_grounded_fluent_with_repeating_parameters_returns_correct_function(
        zenotravel_problem_parser: ProblemParser, zenotravel_domain: Domain):
    valid_grounded_function = ['distance', 'city7', 'city7']
    test_objects_ast = [f"city{i}" for i in range(0, 10)]
    test_objects_ast.extend(['-', 'city'])

    parsed_objects = zenotravel_problem_parser.parse_objects(test_objects_ast)
    zenotravel_problem_parser.problem.objects = parsed_objects

    extracted_function = zenotravel_problem_parser.parse_grounded_numeric_fluent(valid_grounded_function)
    assert extracted_function.name == "distance"
    assert extracted_function.signature == {
        "city7": zenotravel_domain.types["city"]
    }
    assert extracted_function.state_representation == "(= (distance city7 city7) 0)"


def test_parse_parse_grounded_predicate_when_given_wrong_number_of_parameters_raises_an_error(
        problem_parser_with_objects: ProblemParser, domain: Domain):
    invalid_predicate_wrong_num_params = ['next_num', 'num0', 'num1', 'num2']
    lifted_predicate = domain.predicates["next_num"]
    with raises(ValueError):
        problem_parser_with_objects.parse_grounded_predicate(invalid_predicate_wrong_num_params, lifted_predicate)


def test_parse_parse_grounded_predicate_when_given_wrong_parameter_type_raises_error(
        problem_parser_with_objects: ProblemParser, domain: Domain):
    invalid_predicate_wrong_type = ['next_num', 'num0', 'stage4']
    lifted_predicate = domain.predicates["next_num"]
    with raises(AssertionError):
        problem_parser_with_objects.parse_grounded_predicate(invalid_predicate_wrong_type, lifted_predicate)


def test_parse_grounded_predicate_with_legal_predicate_returns_grounded_predicate_object(
        problem_parser_with_objects: ProblemParser, domain: Domain):
    valid_predicate = ['next_num', 'num0', 'num1']
    test_predicate_name = "next_num"
    lifted_predicate = domain.predicates[test_predicate_name]
    grounded_predicate = problem_parser_with_objects.parse_grounded_predicate(valid_predicate, lifted_predicate)
    assert grounded_predicate is not None
    assert grounded_predicate.name == test_predicate_name
    assert grounded_predicate.signature == lifted_predicate.signature
    assert grounded_predicate.object_mapping == {
        "?i1": "num0",
        "?i2": "num1"
    }


def test_parse_grounded_predicate_with_legal_predicate_with_constant_returns_grounded_predicate_object(
        problem_parser_with_objects: ProblemParser, domain: Domain):
    valid_predicate_with_const = ['harvest_phase', 'stage3', 'harvest_end']
    test_predicate_name = "harvest_phase"
    lifted_predicate = domain.predicates[test_predicate_name]
    grounded_predicate = problem_parser_with_objects.parse_grounded_predicate(
        valid_predicate_with_const, lifted_predicate)
    assert grounded_predicate is not None
    assert grounded_predicate.name == test_predicate_name
    assert grounded_predicate.signature == lifted_predicate.signature
    assert grounded_predicate.object_mapping == {
        "?s": "stage3",
        "?hclass": "harvest_end"
    }


def test_parse_state_component_when_given_fluent_component_with_wrong_length_raises_error(
        problem_parser_with_objects: ProblemParser):
    invalid_fluent_state_component = ['=', ['group_worker_cost', 'worker2'], '60', '70', '80']
    with raises(SyntaxError):
        problem_parser_with_objects.parse_state_component(invalid_fluent_state_component)


def test_parse_state_component_with_legal_fluent_set_fluent_value_to_the_input_float_value(
        problem_parser_with_objects: ProblemParser, domain: Domain):
    valid_function_expression = ['=', ['group_worker_cost', 'worker2'], '60']
    problem_parser_with_objects.parse_state_component(valid_function_expression)

    expected_function_str = "(group_worker_cost worker2)"
    assert len(problem_parser_with_objects.problem.initial_state_fluents) == 1
    extracted_function = problem_parser_with_objects.problem.initial_state_fluents[expected_function_str]
    assert extracted_function.value == 60


def test_parse_state_component_when_given_invalid_numeric_term_raises_error(problem_parser_with_objects: ProblemParser):
    invalid_fluent_state_component = ['*', ['group_worker_cost', 'worker2'], '60']
    with raises(ValueError):
        problem_parser_with_objects.parse_state_component(invalid_fluent_state_component)


def test_parse_state_component_when_given_invalid_predicate_term_raises_error(
        problem_parser_with_objects: ProblemParser):
    invalid_fluent_state_component = ['num_substract_blah', 'num3', 'num1', 'num2']
    with raises(ValueError):
        problem_parser_with_objects.parse_state_component(invalid_fluent_state_component)


def test_parse_state_component_with_legal_predicate_set_predicate_in_problem_data(
        problem_parser_with_objects: ProblemParser, domain: Domain):
    valid_predicate = ['next_num', 'num0', 'num1']
    test_predicate_name = "next_num"
    problem_parser_with_objects.parse_state_component(valid_predicate)
    expected_lifted_predicate = domain.predicates[test_predicate_name].untyped_representation
    assert expected_lifted_predicate in problem_parser_with_objects.problem.initial_state_predicates


def test_parse_initial_state_with_single_component_does_not_fail(problem_parser_with_objects: ProblemParser):
    test_initial_state = [['num_substract', 'num1', 'num1', 'num0']]
    try:
        problem_parser_with_objects.parse_initial_state(test_initial_state)
    except Exception:
        fail()


def test_parse_initial_state_with_multiple_component_does_not_fail(problem_parser_with_objects: ProblemParser):
    test_complex_initial_state = [['=', ['group_worker_cost', 'worker2'], '60'],
                                  ['=', ['group_worker_cost', 'worker3'], '30'],
                                  ['=', ['group_worker_cost', 'worker4'], '15'], ['next_num', 'num0', 'num1'],
                                  ['next_num', 'num1', 'num2'], ['next_num', 'num2', 'num3'],
                                  ['next_num', 'num3', 'num4'],
                                  ['next_num', 'num4', 'num5'], ['next_num', 'num5', 'num6'],
                                  ['next_num', 'num6', 'num7'],
                                  ['next_num', 'num7', 'num8'], ['next_num', 'num8', 'num9'],
                                  ['next_num', 'num9', 'num10'],
                                  ['next_num', 'num10', 'num11'], ['next_num', 'num11', 'num12'],
                                  ['next_num', 'num12', 'num13'], ['next_num', 'num13', 'num14'],
                                  ['next_num', 'num14', 'num15'], ['next_num', 'num15', 'num16'],
                                  ['num_substract', 'num1', 'num1', 'num0'], ['num_substract', 'num2', 'num1', 'num1'],
                                  ['num_substract', 'num2', 'num2', 'num0'], ['num_substract', 'num3', 'num1', 'num2'],
                                  ['num_substract', 'num3', 'num2', 'num1'], ['num_substract', 'num3', 'num3', 'num0'],
                                  ['num_substract', 'num4', 'num1', 'num3'], ['num_substract', 'num4', 'num2', 'num2'],
                                  ['num_substract', 'num4', 'num3', 'num1'], ['num_substract', 'num4', 'num4', 'num0'],
                                  ['num_substract', 'num5', 'num1', 'num4'], ['num_substract', 'num5', 'num2', 'num3'],
                                  ['num_substract', 'num5', 'num3', 'num2'], ['num_substract', 'num5', 'num4', 'num1'],
                                  ['num_substract', 'num5', 'num5', 'num0'], ['num_substract', 'num6', 'num1', 'num5'],
                                  ['num_substract', 'num6', 'num2', 'num4'], ['num_substract', 'num6', 'num3', 'num3'],
                                  ['num_substract', 'num6', 'num4', 'num2'], ['num_substract', 'num6', 'num5', 'num1'],
                                  ['num_substract', 'num6', 'num6', 'num0'], ['num_substract', 'num7', 'num1', 'num6'],
                                  ['num_substract', 'num7', 'num2', 'num5'], ['num_substract', 'num7', 'num3', 'num4'],
                                  ['num_substract', 'num7', 'num4', 'num3'], ['num_substract', 'num7', 'num5', 'num2'],
                                  ['num_substract', 'num7', 'num6', 'num1'], ['num_substract', 'num7', 'num7', 'num0'],
                                  ['num_substract', 'num8', 'num1', 'num7'], ['num_substract', 'num8', 'num2', 'num6'],
                                  ['num_substract', 'num8', 'num3', 'num5'], ['num_substract', 'num8', 'num4', 'num4'],
                                  ['num_substract', 'num8', 'num5', 'num3'], ['num_substract', 'num8', 'num6', 'num2'],
                                  ['num_substract', 'num8', 'num7', 'num1'], ['num_substract', 'num8', 'num8', 'num0'],
                                  ['num_substract', 'num9', 'num1', 'num8'], ['num_substract', 'num9', 'num2', 'num7'],
                                  ['num_substract', 'num9', 'num3', 'num6'], ['num_substract', 'num9', 'num4', 'num5'],
                                  ['num_substract', 'num9', 'num5', 'num4']]
    try:
        problem_parser_with_objects.parse_initial_state(test_complex_initial_state)
    except Exception:
        fail()


def test_parse_goal_state_when_parsing_ast_that_does_not_start_with_and_raises_error(
        problem_parser_with_objects: ProblemParser):
    bad_goal_state = ['blah', ['harvest_phase', 'stage3', 'harvest_end']]
    with raises(SyntaxError):
        problem_parser_with_objects.parse_goal_state(bad_goal_state)


def test_parse_goal_state_when_parsing_simple_numeric_condition_the_numeric_tree_is_created_correctly(
        problem_parser_with_objects: ProblemParser):
    goal_state_numeric_tokens = PDDLTokenizer(
        pddl_str="(and (>= (group_worker_cost worker1) 1))").parse()
    problem_parser_with_objects.parse_goal_state(goal_state_numeric_tokens)
    assert len(problem_parser_with_objects.problem.goal_state_fluents) == 1
    for fluent in problem_parser_with_objects.problem.goal_state_fluents:
        print(fluent)


def test_parse_goal_state_when_parsing_complex_numeric_condition_the_numeric_tree_is_created_correctly(
        problem_parser_with_objects: ProblemParser):
    goal_state_numeric_tokens = PDDLTokenizer(
        pddl_str="(and (>= (+ (* 2.0 (group_worker_cost worker0)) "
                 "(+ (* 1.0 (group_worker_cost worker1)) (+ (* 1.5 (group_worker_cost worker2)) 0))) 14.0))").parse()
    problem_parser_with_objects.parse_goal_state(goal_state_numeric_tokens)
    assert len(problem_parser_with_objects.problem.goal_state_fluents) == 1
    for fluent in problem_parser_with_objects.problem.goal_state_fluents:
        print(fluent)


def test_parse_goal_state_when_parsing_multiple_numeric_conditions_the_numeric_tree_is_created_correctly(
        problem_parser_with_objects: ProblemParser):
    goal_state_numeric_tokens = PDDLTokenizer(
        pddl_str="""(and
			(>= (group_worker_cost worker0) 1)
			(>= (group_worker_cost worker1) 1)
			(>= (group_worker_cost worker2) 1)
			(>= (+ (* 2.0 (group_worker_cost worker0))(+ (* 1.0 (group_worker_cost worker1))(+ (* 1.5 (group_worker_cost worker2)) 0))) 14.0)
		)""").parse()

    problem_parser_with_objects.parse_goal_state(goal_state_numeric_tokens)
    assert len(problem_parser_with_objects.problem.goal_state_fluents) == 4
    for fluent in problem_parser_with_objects.problem.goal_state_fluents:
        print(fluent)


def test_parse_goal_state_with_legal_goal_state_extract_correct_number_of_grounded_predicates(
        problem_parser_with_objects: ProblemParser):
    valid_goal_state = ['and', ['harvest_phase', 'stage3', 'harvest_end']]
    problem_parser_with_objects.parse_goal_state(valid_goal_state)
    assert len(problem_parser_with_objects.problem.goal_state_predicates) == 1


def test_parse_problem_with_real_problem_data_does_not_fail(problem_parser):
    try:
        problem = problem_parser.parse_problem()
        print(problem)
    except Exception:
        fail()



import enum
import random
import sys
from collections import OrderedDict
from queue import Queue


class typeStatus(enum.Enum):
    OFF = 0
    ON = 1


typing = typeStatus
st = str()

class probTypes(enum.Enum):
    STRIPS = 0
    NUMERIC = 1
    SIMPLETIME = 2
    TIMED = 3
    COMPLEX = 4
    HARDNUMERIC = 5


probType = probTypes

def rnd(limit=-1):
    if limit == -1:
        return float(random.random())
    return int(limit * random.random())

class ProblemObject:
    class OutStatus(enum.Enum):
        OBJECT = 0
        INIT = 1
        GOAL = 2
    def object(self):
        raise NotImplementedError
    def init(self):
        raise NotImplementedError
    def goal(self):
        raise NotImplementedError
    def write(self):
        global outStatus
        if outStatus == ProblemObject.OutStatus.OBJECT:
            self.object()
        elif outStatus == ProblemObject.OutStatus.INIT:
            self.init()
        elif outStatus == ProblemObject.OutStatus.GOAL:
            self.goal()
    @classmethod
    def Objects(self):
        global outStatus
        outStatus = ProblemObject.OutStatus.OBJECT
    @classmethod
    def Inits(self):
        global outStatus
        outStatus = ProblemObject.OutStatus.INIT
    @classmethod
    def Goals(self):
        global outStatus
        outStatus = ProblemObject.OutStatus.GOAL
    
outStatus = ProblemObject.OutStatus.OBJECT

class IOManipulator(object):
    def __init__(self, function=None):
        self.function = function
    def do(self, output):
        self.function(output)
def do_endl(stream):
    stream.output.write('\n')
endl = IOManipulator(do_endl)

class OStream(object):
    def __init__(self, output=None):
        if output is None:
            output = sys.stdout
        self.output = output
        self.format = '%s'
    def __lshift__(self, thing):
        if isinstance(thing, IOManipulator):
            self.output.write(thing)
            return OStream()
        else:
            self.output.write(self.format % thing)
            self.format = '%s'
        return OStream()


seaZones = [set() for _ in range(4)]

id = 0

class Place(ProblemObject):
	
    def __init__(self):
        global id
        self.__place = id
        id += 1
        self.__mountain = rnd(10) < 3
        self.__woodland = rnd(10) < 6
        self.__metalliferous = rnd(10) < 2
        self.__bycoast = rnd(10) < 5
        self.__seaZone = rnd(4)
        global seaZones
        if self.__bycoast: 
            seaZones[self.__seaZone].add(self.__place)

    def makeMountain(self):
        self.__mountain = True
    def makeWoodland(self):
        self.__woodland = True
    def makeMetalliferous(self):
        self.__metalliferous = True

    def object(self):
        global typing
        global st
        st += f"\tlocation{self.__place}"
        if typing == typeStatus.ON:
            st += " - place"
        st += "\n"
    def init(self):
        global typing
        global st
        if typing == typeStatus.OFF:
            st += f"\t(place location{self.__place})\n"
        if self.__mountain:
            st += f"\t(mountain location{self.__place})\n"
        if self.__woodland:
            st += f"\t(woodland location{self.__place})\n"
        if self.__bycoast:
            st += f"\t(by-coast location{self.__place})\n"
        if self.__metalliferous:
            st += f"\t(metalliferous location{self.__place})\n"
        st += f"\t(= (housing location{self.__place}) 0)\n" + \
            f"\t(= (available wood location{self.__place}) 0)\n" + \
            f"\t(= (available timber location{self.__place}) 0)\n" + \
            f"\t(= (available ore location{self.__place}) 0)\n" + \
            f"\t(= (available stone location{self.__place}) 0)\n" + \
            f"\t(= (available iron location{self.__place}) 0)\n" + \
            f"\t(= (available coal location{self.__place}) 0)\n"
    def goal(self):
        pass
    def __str__(self):
        return st


class Goal(ProblemObject):
    def __del__(self):
        pass
    def object(self):
        raise NotImplementedError
    def init(self):
        raise NotImplementedError

def random_select(s):
    i = rnd(len(s))
    for val in s:
        while i > 0:
            i-=1
        return val
    return val

links = set()

class RailLink(Goal):
    def __init__(self, g, nlocs, length):
        self.start = rnd(nlocs)
        self.__locs = list()
        self.__locs.append(self.start)
        for i in range(length):
            if not g[self.start]: return # graph g int -> set
            self.start = random_select(g[self.start])
            if self.__locs[len(self.__locs) - 1] != self.start:
                self.__locs.append(self.start)
    def goal(self):
        global st
        global links
        for i in range(len(self.__locs) - 1):
            if (self.__locs[i], self.__locs[i+1]) not in links:
                st += f"\t(connected-by-rail location{self.__locs[i]} location{self.__locs[i+1]})\n"
                links.add((self.__locs[i], self.__locs[i+1]))
    def __repr__(self):
        return st

allBuildings = set()
numBuildings = 3
buildingName = ["coal-stack", "sawmill", "ironworks"]

class Building(Goal):
    def __init__(self, nlocs):
        global allBuildings
        global numBuildings
        global buildingName
        self.__building = rnd(numBuildings)
        self.__loc = rnd(nlocs) 
        while self.__loc < nlocs:
            while self in allBuildings and self.__building < numBuildings:
                self.__building+=1
                if self.__building < numBuildings:
                    break
            self.__loc+=1
            self.__building = 0
        allBuildings.add(self)
    def __hash__(self):
        return hash((self.__building, self.__loc))
    def __eq__(self, b):
        return self.__building == b.__building and self.__loc == b.__loc
    def __lt__(self, b):
        return self.__loc < b.loc or (self.__loc == b.loc and self.__building < b.building)
    def goal(self):
        global buildingName
        global st
        st += f"\t(has-{buildingName[self.__building]} location{self.__loc})\n"
    def __repr__(self):
        return st

places = set()

class Housing(Goal):
    def __init__(self, nlocs):
        global places
        if not places:
            for i in range(nlocs):
                places.add(i)
        self.__place = random_select(places)
        self.__quantity = rnd(2) + 1
        places.remove(self.__place)
    def goal(self):
        global st
        st += f"\t(>= (housing location{self.__place}) {self.__quantity})\n"
    def __repr__(self):
        return st


class Map(ProblemObject):
    def __explore(self, g, start, reached):
        togo = Queue()
        togo.put(start)
        reached.add(start) #set
        while not togo.empty():
            loc = togo.get()
            for i in g[loc]:
                if i not in reached:
                    togo.put(i)
                    reached.add(i)

    def __connect(self, g): 
        reached = set()
        start = rnd(self.__places)
        self.__explore(g, start, reached)
        while len(reached) != self.__places:
            next = int()
            for i in range(self.__places):
                if i not in reached:
                    next = i
                    break
            g[start].add(next)
            start = next
            self.__explore(g, start, reached)
    def __init__(self, s, ngs): 
        self.__places = s
        self.__Places = list()
        self.__road = OrderedDict()
        self.__goals = list()
        for _ in range(s):
            self.__Places.append(Place())
            for _ in range(4):
                f = rnd(s)
                t = rnd(s)
                if t not in self.__road:
                    self.__road[t] = set()
                if f not in self.__road[t]:
                    if f not in self.__road:
                        self.__road[f] = set()
                    self.__road[f].add(t)
        self.__connect(self.__road)
        self.__Places[rnd(self.__places)].makeMountain()
        self.__Places[rnd(self.__places)].makeWoodland()
        self.__Places[rnd(self.__places)].makeMetalliferous()

        for _ in range(ngs):
            if rnd(10) < 3:
                self.__goals.append(RailLink(self.__road, self.__places, int(ngs/3)))
            else:
                if rnd(10) < 3:
                    self.__goals.append(Housing(self.__places))
                else:
                    self.__goals.append(Building(self.__places))
    def __del__(self):
        for i in self.__goals:
            del i

    def object(self):
        global st
        for p in self.__Places:
            p.object()
    def init(self):
        global seaZones
        global typing
        global st
        for p in self.__Places:
            p.init()
        if typing == typeStatus.OFF:
            st += "\t(resource iron)\n\t(resource wood)\n\t(resource timber)\n\t(resource ore)\n\t(resource stone)\n\t(resource coal)\n"
		
        for k, v in self.__road.items():
            for ele in v:
                if k == ele:
                    continue
                if ele in self.__road:
                    if ele in self.__road:
                        if k in self.__road[ele]:
                            continue

                st += f"\t(connected-by-land location{k} location{ele})\n\t(connected-by-land location{ele} location{k})\n"
        for i in range(4):
            for idx, s in enumerate(seaZones[i]):
                for t in [t1 for index, t1 in enumerate(seaZones[i]) if index > idx]:
                    st += f"\t(connected-by-sea location{s} location{t})\n\t(connected-by-sea location{t} location{s})\n"

    def goal(self):
        global st
        for i in self.__goals:
            i.goal()
    def __repr__(self):
        return st

ID = 0

class Vehicle(ProblemObject):
    def __init__(self):
        global ID
        self.__vnum = ID + 1
        ID +=1
    def object(self):
        global st
        global typing
        st += f"\tvehicle{self.__vnum}"
        if typing == typeStatus.ON:
            st += " - vehicle"
        st += "\n"
    def init(self):
        global typing
        global st
        if typing == typeStatus.OFF:
            st += f"\t(vehicle vehicle{self.__vnum})\n"
        st += f"\t(potential vehicle{self.__vnum})\n"
    def goal(self):
        pass
    def __repr__(self):
        return st

def main(argc, argv):
    global typing
    global probType
    typing = typeStatus.ON;
    probType = probTypes.STRIPS;
    global o
    global st
    if argc < 4:
        print("Usage: settlers [-u] <seed> #<locations> #<goals> #<vehicles>\n\n\tThe number of vehicles is the number of \"potential\" vehicles.\n\n\t-u: Untyped\n\n")
        exit(0)
    argv = argv[1:]
    argc -= 1
    if argv[0][0] == '-':
        if argv[0][1] == 'u': 
            typing = typeStatus.OFF
        argv = argv[1:]
        argc -= 1
    random.seed(int(argv[0]))
    p = Map(int(argv[1]),int(argv[2]))
    carts = list()
    for _ in range(int(argv[3])):
        carts.append(Vehicle())
    print("(define (problem settlers)\n(:domain civ)\n(:objects\n",end='')
    p.write()
    o << st
    st = ''
    for c in carts:
        c.write()
        o << st
        st = ''
    print(")\n(:init\n\t(= (resource-use) 0)\n\t(= (labour) 0)\n\t(= (pollution) 0)\n",end='')
    ProblemObject.Inits()
    p.write()
    o << st
    st = ''
    for c in carts:
        c.write()
        o << st
        st = ''
    print(")\n(:goal (and\n",end='')
    ProblemObject.Goals()
    p.write()
    o << st
    st = ''
    print(f"\t)\n)\n\n(:metric minimize (+ (+ (* {rnd(4)} (pollution)) (* {rnd(4)} (resource-use))) (* {rnd(4)} (labour))))\n)\n",end='')

o = OStream()

if __name__ == '__main__':
    argv = sys.argv
    argc = len(argv)
    main(argc, argv)




"""Module that represents a PDDL+ action."""
from typing import Set, List, Tuple

from .universal_quantifier import UniversalQuantifiedEffect
from .conditional_effect import ConditionalEffect
from .numerical_expression import NumericalExpressionTree
from .pddl_predicate import SignatureType, Predicate


class Action:
    """Class representing an instantaneous action in a PDDL+ problems."""

    name: str
    signature: SignatureType
    positive_preconditions: Set[Predicate]
    negative_preconditions: Set[Predicate]
    numeric_preconditions: Set[NumericalExpressionTree]
    equality_preconditions: Set[Tuple[str, str]]
    inequality_preconditions: Set[Tuple[str, str]]
    add_effects: Set[Predicate]
    delete_effects: Set[Predicate]
    conditional_effects: Set[ConditionalEffect]
    universal_effects: Set[UniversalQuantifiedEffect]
    numeric_effects: Set[NumericalExpressionTree]
    # currently only supporting disjunction of numeric preconditions
    disjunctive_numeric_preconditions: List[Set[NumericalExpressionTree]]

    def __init__(self):
        self.positive_preconditions = set()
        self.negative_preconditions = set()
        self.numeric_preconditions = set()
        self.equality_preconditions = set()
        self.inequality_preconditions = set()
        self.add_effects = set()
        self.delete_effects = set()
        self.numeric_effects = set()
        self.disjunctive_numeric_preconditions = []
        self.conditional_effects = set()
        self.universal_effects = set()

    def __str__(self):
        signature_str_items = []
        for parameter_name, parameter_type in self.signature.items():
            signature_str_items.append(f"{parameter_name} - {str(parameter_type)}")

        signature_str = " ".join(signature_str_items)
        return f"({self.name} {signature_str})"

    @property
    def parameter_names(self) -> List[str]:
        return list(self.signature.keys())



"""Module to convert MA problems into single agent domains"""
import logging
from pathlib import Path

from pddl_plus_parser.exporters import ProblemExporter
from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser
from pddl_plus_parser.models import Problem


class MultiAgentProblemsConverter:
    """Converts factored multi-agent problems to single agent problems."""

    logger: logging.Logger
    problems_directory_path: Path

    def __init__(self, working_directory_path: Path, problem_file_prefix: str):
        self.logger = logging.getLogger(__name__)
        self.problems_directory_path = working_directory_path
        self.problem_file_prefix = problem_file_prefix

    def combine_problems(self, combined_domain_path: Path) -> Problem:
        """Converts the MA problems to one single agent problem with combined initial state and goals.

        :return: the problem that represents the combination of all the agents' problems.
        """
        self.logger.info("Starting to combine the problems")
        combined_domain = DomainParser(domain_path=combined_domain_path, partial_parsing=False).parse_domain()
        combined_problem = Problem(domain=combined_domain)
        for problem_file_path in self.problems_directory_path.glob(f"{self.problem_file_prefix}-*.pddl"):
            agent_problem = ProblemParser(problem_path=problem_file_path, domain=combined_domain).parse_problem()
            combined_problem.name = agent_problem.name
            combined_problem.objects.update(agent_problem.objects)
            combined_problem.initial_state_fluents.update(agent_problem.initial_state_fluents)
            for predicate, grounded_predicates in agent_problem.initial_state_predicates.items():
                combined_state_predicates = [p.untyped_representation for p in
                                             combined_problem.initial_state_predicates[predicate]]
                for grounded_predicate in grounded_predicates:
                    if grounded_predicate.untyped_representation in combined_state_predicates:
                        continue

                    combined_problem.initial_state_predicates[predicate].add(grounded_predicate)

            combined_problem.goal_state_predicates.extend(agent_problem.goal_state_predicates)
            combined_problem.goal_state_predicates = list(set(combined_problem.goal_state_predicates))
            combined_problem.goal_state_fluents.update(agent_problem.goal_state_fluents)

        return combined_problem

    def export_combined_problem(self, combined_domain_path: Path) -> None:
        """Export the combined multi-agent problem to a single PDDL file.

        :param combined_domain_path: the path to the combined domain (containing the agents' data).
        """
        combined_problem = self.combine_problems(combined_domain_path)
        ProblemExporter().export_problem(combined_problem, self.problems_directory_path / f"combined_problem.pddl")



#!/usr/bin/python3
import argparse
import itertools
import random
from pathlib import Path

from pddl_plus_parser.problem_generators import get_problem_template

TEMPLATE_FILE_PATH = Path("plant_watering_template.pddl")


def construct_pouring_inner_goal(num_plants) -> str:
    """

    :param num_plants:
    :return:
    """
    if num_plants == 0:
        return f"(poured plant{num_plants})"

    inner_layer = construct_pouring_inner_goal(num_plants - 1)
    return f"(+ (poured plant{num_plants}) {inner_layer})"


def generate_instance(instance_name: str, num_taps: int, num_agents: int, num_plants: int) -> str:
    template = get_problem_template(TEMPLATE_FILE_PATH)
    max_int = 10 * 2 * num_plants
    template_mapping = {
        "instance_name": instance_name,
        "domain_name": "mt-plant-watering-constrained",
        "taps_list": " ".join([f"tap{i}" for i in range(num_taps)]),
        "agents_list": " ".join([f"agent{i}" for i in range(num_agents)]),
        "plants_list": " ".join([f"plant{i}" for i in range(num_plants)]),
        "max_int_value": str(max_int),
        "poured_plants": "\n".join([f"(= (poured plant{i}) 0)" for i in range(num_plants)]),
        "agents_locations": "\n".join([f"(= (x agent{i}) {coordinate})\n\t\t(= (y agent{i}) {coordinate})"
                                       for i, coordinate in zip(range(num_agents),
                                                                random.choices(range(10), k=num_agents))]),
        "plants_locations": "\n".join([f"(= (x plant{i}) {coordinate})\n\t\t(= (y plant{i}) {coordinate})"
                                       for i, coordinate in zip(range(num_agents),
                                                                random.choices(range(10), k=num_plants))]),
        "taps_locations": "\n".join([f"(= (x tap{i}) {coordinate})\n\t\t(= (y tap{i}) {coordinate})"
                                     for i, coordinate in zip(range(num_agents),
                                                              random.choices(range(10), k=num_taps))]),
        "pour_goals": "\n".join([f"(= (poured plant{i}) {val})" for i, val in zip(range(num_agents),
                                                                                  random.choices(range(max_int),
                                                                                                 k=num_plants))]),
        "total_poured_goal": f"(= (total_poured) {construct_pouring_inner_goal(num_plants)} )",
    }
    return template.substitute(template_mapping)


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Generate plat watering instance")
    parser.add_argument("--num_taps", required=True, help="taps to plant the water with")
    parser.add_argument("--num_agents", required=True, help="number of agents that can water the plants")
    parser.add_argument("--num_plants", required=True, help="the number of plants to water")
    args = parser.parse_args()
    return args


def generate_multiple_problems(
        num_taps: int, num_agents: int, num_plants: int, output_folder: Path):
    for i in range(1, num_taps + 1):
        for j in range(1, num_agents + 1):
            for k in range(1, num_plants + 1):
                instance_name = f"{i}_{j}_{k}"
                with open(output_folder / f"pfile{instance_name}_{random.randint(0, 100)}.pddl", "wt") as problem_file:
                    problem_file.write(generate_instance(f"instance_{instance_name}", i, j, k))


def main():
    generate_multiple_problems(10, 10, 10, Path("C:\Argaman\Planning\Minecraft\more_domains\plant_watering"))


if __name__ == '__main__':
    main()



# To change this license    header, choose License Headers in Project Properties.
# To change this template file, choose Tools | Templates
# and open the template in the editor.
# !/usr/bin/python3
import argparse
import itertools
import random
from enum import Enum
from pathlib import Path
from typing import NoReturn

import networkx as nx

from pddl_plus_parser.problem_generators import get_problem_template

TEMPLATE_FILE_PATH = Path("farmland_template.pddl")


class GraphGeneratorTypes(Enum):
    star = 1
    strogaz = 2


def generate_adjacent_graph(graph_generator: GraphGeneratorTypes, num_farms: int) -> nx.Graph:
    """Generates the graph of farms where there are adjacent farms represented in a graph.

    :param graph_generator: the type of graph generator to use.
    :param num_farms: the number of farms in the graph.
    :return: the graph of farms.
    """
    if graph_generator == GraphGeneratorTypes.star:
        G = nx.star_graph(num_farms - 1)

    elif graph_generator == GraphGeneratorTypes.strogaz:
        G = nx.connected_watts_strogatz_graph(num_farms, k=min(2, num_farms - 1), tries=10000, p=0.5)

    else:
        G = nx.ladder_graph(num_farms // 2)
    return G


def generate_instance(
        instance_name: str, num_farms: int, num_units: int, graph_generator: GraphGeneratorTypes) -> str:
    """Generates the farmland planning problem instance.

    :param instance_name: the name of the problem instance.
    :param num_farms: the number of farms in the problem.
    :param num_units: the number of units in the problem.
    :param graph_generator: the type of generator to use to generate the adjacent graph.
    :return: the string representation of the problem instance.
    """
    template = get_problem_template(TEMPLATE_FILE_PATH)
    template_mapping = {"instance_name": instance_name, "domain_name": "farmland"}

    G = generate_adjacent_graph(graph_generator, num_farms)

    source = random.randint(0, num_farms - 1)
    farms_init_fluents = []
    farms_reward_bound = ["(>= "]
    farms_goal_fluents = []
    adjacents = []
    farms = " ".join([f"farm{i} " for i in range(num_farms)])
    for i in range(num_farms):
        weight_bound = "{0:.1f}".format(random.random() + 1.0) if i != source else 1.0
        init_fluent = f"(= (x farm{i}) {str(random.randint(0, 1))})\n\t\t" if i != source else \
            f"(= (x farm{i}) {num_units})\n\t\t"
        farms_init_fluents.append(init_fluent)
        farms_reward_bound.append(f"(+ (* {weight_bound} (x farm{i}))")
        farms_goal_fluents.append(f"(>= (x farm{i}) 1)\n\t\t\t")

        for element in G[i]:
            adjacents.append(f"(adj farm{i} farm{str(element)})\n\t\t")

    farms_reward_bound.append(" 0")
    farms_reward_bound.append(")" * num_farms)
    farms_reward_bound.append(f" {str(num_units * 1.4)})")

    template_mapping['farm_name_list'] = farms
    template_mapping['farm_init_allocation'] = "".join(farms_init_fluents)
    template_mapping['farm_connections'] = "".join(adjacents)
    template_mapping['farm_final_requirement'] = "".join(farms_goal_fluents)
    template_mapping['overall_reward_bound'] = "".join(farms_reward_bound)

    return template.substitute(template_mapping)


def parse_arguments() -> argparse.Namespace:
    """Parses the command line arguments."""
    parser = argparse.ArgumentParser(description="Generate farmland planning instance")
    parser.add_argument("--min_farms", required=True, help="The minimal number of farms possible in a planning problem")
    parser.add_argument("--max_farms", required=True, help="The maximal number of farms possible in a planning problem")
    parser.add_argument("--min_num_units", required=True, help="Minimal number of units")
    parser.add_argument("--max_num_units", required=True, help="Maximal number of units")
    parser.add_argument("--output_folder", required=True, help="The path to the output folder")
    parser.add_argument("--graph_generator", required=False,
                        help="Graph Generator between star (default) or strogatz or ladder",
                        default=GraphGeneratorTypes.star)
    return parser.parse_args()


def generate_multiple_problems(min_farms: int, max_farms: int, min_num_units: int, max_num_units: int,
                               output_folder: Path,
                               graph_generator: GraphGeneratorTypes = GraphGeneratorTypes.star) -> NoReturn:
    """Generate multiple problems based on the input arguments.

    :param min_farms: the minimal number of farms possible in a planning problem.
    :param max_farms: the maximal number of farms possible in a planning problem.
    :param min_num_units: the minimal number of units.
    :param max_num_units: the maximal number of units.
    :param output_folder: the path to the output folder where the planning problems will be saved.
    :param graph_generator: the type of graph generator to use.
    """
    farms_range = [i for i in range(min_farms, max_farms + 1)]
    units_range = [i for i in range(min_num_units, max_num_units + 1)]
    for num_farms, num_units in itertools.product(farms_range, units_range):
        print(f"Generating problem with {num_farms} farms and {num_units} units")
        with open(output_folder / f"pfile{num_farms}_{num_units}.pddl", "wt") as problem_file:
            problem_file.write(generate_instance(f"instance_{num_farms}_{num_units}", num_farms, num_units,
                                                 graph_generator))


def main():
    args = parse_arguments()
    generate_multiple_problems(min_farms=int(args.min_farms),
                               max_farms=int(args.max_farms),
                               min_num_units=int(args.min_num_units),
                               max_num_units=int(args.max_num_units),
                               output_folder=Path(args.output_folder),
                               graph_generator=args.graph_generator)


if __name__ == '__main__':
    main()
